import pathlib

import pandas as pd
import pytest
import yaml

import zntrack
from zntrack.abc import Metrics

# from zntrack.utils import config, run_dvc_cmd # TODO ADD BACK
# TODO ADD ZNTRACK PLOTS

run_dvc_cmd = None
config = None


class NodeWithPlotsDVC(zntrack.Node):
    data: pd.DataFrame = zntrack.plots(
        zntrack.nwd / "data.csv",
        x="x",
        y="y",
        x_label="x",
        y_label="y",
        title="title",
        template="linear",
        use_global_plots=False,
    )

    def run(self) -> None:
        df = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]})
        df.to_csv(self.data)

    def get_data(self):
        return pd.read_csv(self.data)


class NodeWithPlotsZn(zntrack.Node):
    data: pd.DataFrame = zntrack.plots(
        x="x",
        y="y",
        x_label="x",
        y_label="y",
        title="title",
        template="linear",
        use_global_plots=False,
    )

    def run(self) -> None:
        self.data = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]})

    def get_data(self):
        return self.data


class NodeWithPlotsZnGlobal(zntrack.Node):
    data: pd.DataFrame = zntrack.plots(
        x="x",
        y=["y", "z"],
        x_label="x",
        y_label="y",
        title="title",
        template="linear",
    )
    metrics: Metrics = zntrack.metrics()

    def run(self) -> None:
        self.data = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6], "z": [7, 8, 9]})
        self.metrics = {"a": 1, "b": 2}

    def get_data(self):
        return self.data


class NodeWithPlotsDVCGlobal(zntrack.Node):
    data: pd.DataFrame = zntrack.plots(
        zntrack.nwd / "data.csv",
        x="x",
        y=["y", "z"],
        x_label="x",
        y_label="y",
        title="title",
        template="linear",
    )
    metrics: Metrics = zntrack.metrics(zntrack.nwd / "metrics.json")

    def run(self) -> None:
        df = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6], "z": [7, 8, 9]})
        df.to_csv(self.data)
        self.metrics.write_text("{'a': 1, 'b': 2}")

    def get_data(self):
        return pd.read_csv(self.data)


class NodeRemoteTemplate(zntrack.Node):
    data: pd.DataFrame = zntrack.plots(
        zntrack.nwd / "data.csv",
        template=zntrack.__file__,
    )

    def run(self) -> None:
        df = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6], "z": [7, 8, 9]})
        df.to_csv(self.data)


@pytest.mark.parametrize("cls", [NodeWithPlotsDVC, NodeWithPlotsZn])
@pytest.mark.parametrize("eager", [True, False])
def test_NodeWithPlots(proj_path, eager, cls):
    with zntrack.Project() as project:
        node = cls()
    project.run(eager=eager)

    if not eager:
        node.load()
        dvc_dict = yaml.safe_load((proj_path / config.files.dvc).read_text())

        plots = {
            "x": "x",
            "y": "y",
            "x_label": "x",
            "y_label": "y",
            "title": "title",
            "template": "linear",
        }

        assert (
            dvc_dict["stages"][node.name]["plots"][0][f"nodes/{node.name}/data.csv"]
            == plots
        )
        run_dvc_cmd(["plots", "show"])

    if isinstance(node, NodeWithPlotsDVC):
        assert NodeWithPlotsDVC.from_rev().data == pathlib.Path(
            "nodes", node.name, "data.csv"
        )

    assert node.get_data()["x"].tolist() == [1, 2, 3]


@pytest.mark.parametrize("cls", [NodeWithPlotsZnGlobal, NodeWithPlotsDVCGlobal])
@pytest.mark.parametrize("eager", [True, False])
def test_NodeWithPlotsGlobal(proj_path, eager, cls):
    with zntrack.Project() as project:
        node = cls()
    project.run(eager=eager)

    if not eager:
        node.load()
        dvc_dict = yaml.safe_load((proj_path / config.files.dvc).read_text())

        plots = {
            "x": "x",
            "y": ["y", "z"],
            "x_label": "x",
            "y_label": "y",
            "title": "title",
            "template": "linear",
        }

        assert dvc_dict["plots"][0][f"nodes/{node.name}/data.csv"] == plots

        run_dvc_cmd(["plots", "show"])

    assert node.get_data()["x"].tolist() == [1, 2, 3]


def test_multiple_plots_nodes(proj_path):
    with zntrack.Project(automatic_node_names=True) as project:
        NodeWithPlotsZnGlobal()
        NodeWithPlotsZnGlobal()
        NodeWithPlotsDVCGlobal()
        a = NodeWithPlotsDVCGlobal()
    project.run()
    run_dvc_cmd(["plots", "show"])

    with project:
        NodeWithPlotsZnGlobal()
        b = NodeWithPlotsZnGlobal()
        NodeWithPlotsDVCGlobal()
        NodeWithPlotsDVCGlobal()
    project.run()
    run_dvc_cmd(["plots", "show"])
    # run_dvc_cmd(["repro", "-f"])

    # check loading
    NodeWithPlotsDVCGlobal.from_rev(name=a.name)
    NodeWithPlotsZnGlobal.from_rev(name=b.name)


def test_NodeRemoteTemplate(proj_path):
    with zntrack.Project(automatic_node_names=True) as project:
        NodeRemoteTemplate()
        NodeRemoteTemplate()
    project.run(repro=False)

    assert pathlib.Path("dvc_plots", "templates", "__init__.py").exists()



from zntrack.fields.base import field
from zntrack.fields.deps import deps
from zntrack.fields.outs_and_metrics import metrics, outs
from zntrack.fields.params import params
from zntrack.fields.plots import plots
from zntrack.fields.x_path import (
    deps_path,
    metrics_path,
    outs_path,
    params_path,
    plots_path,
)

# TODO: default file names like `nwd/metrics.json`,
#  `nwd/node-meta.json`, `nwd/plots.csv` should raise
#  an error if passed to `metrics_path` etc.
# TODO: zntrack.outs() and zntrack.outs(cache=False) needs different files!


__all__ = [
    "outs_path",
    "params_path",
    "plots_path",
    "metrics_path",
    "deps_path",
    "deps",
    "params",
    "plots",
    "metrics",
    "outs",
    "field",
]



import pandas as pd
import pandas.testing as pdt
import pytest

import zntrack.examples
from zntrack.config import NodeStatusEnum


@pytest.mark.parametrize("eager", [True, False])
def test_WritePlots(proj_path, eager):
    with zntrack.Project() as project:
        plots = zntrack.examples.WritePlots()

    assert plots.state.state == NodeStatusEnum.CREATED

    if eager:
        project.run()
    else:
        project.repro()
    pdt.assert_frame_equal(plots.plots, pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]}))



import git
import pytest

import zntrack.examples


@pytest.mark.xfail(reason="pending implementation")
@pytest.mark.parametrize("automatic_node_names", [True, False])
def test_multiple_experiments(proj_path, automatic_node_names):
    with zntrack.Project(automatic_node_names=automatic_node_names) as proj:
        node = zntrack.examples.ParamsToOuts(1)
    proj.build()
    proj.repro()

    # node.load()
    # assert node.outputs == 1

    repo = git.Repo()
    # add and commit all files
    repo.git.add(A=True)
    repo.index.commit("commit1")

    for exp in range(5):
        with proj.create_experiment(f"exp{exp}", queue=False):
            node.params = exp

    nodes = [zntrack.examples.ParamsToOuts.from_rev(rev=f"exp{exp}") for exp in range(5)]
    for idx, node in enumerate(nodes):
        assert node.outs == idx
        assert node._external_

    with proj:  # test with / without automatic node names
        node2 = zntrack.examples.AddNodeNumbers(nodes)

    proj.run()
    node2.load()

    assert node2.sum == 10



import pathlib

import pytest

import zntrack


class OutsTmp(zntrack.Node):
    file: pathlib.Path = zntrack.outs_path(zntrack.nwd / "file.txt")
    path: pathlib.Path = zntrack.outs_path(zntrack.nwd / "path")

    def run(self):
        self.path.mkdir(parents=True, exist_ok=True)

        self.file.write_text("Hello, World!")
        (self.path / "data.txt").write_text("Lorem Ipsum")


def test_outs_tmp(proj_path):
    project = zntrack.Project()
    with project:
        node = OutsTmp()

    project.repro()
    assert node.nwd == pathlib.Path("nodes", OutsTmp.__name__)
    with pytest.warns(
        UserWarning,
        match="The temporary path is not used when neither remote or rev are set.",
    ):
        with node.state.use_tmp_path() as path:
            assert node.nwd == pathlib.Path("nodes", OutsTmp.__name__)

    # temp path is not used when neither remote or rev are set.
    node = node.from_rev(remote=".")

    assert node.nwd == pathlib.Path("nodes", OutsTmp.__name__)

    with node.state.use_tmp_path() as path:
        assert node.nwd != pathlib.Path("nodes", OutsTmp.__name__)
        assert node.nwd == path
        assert node.file == path / "file.txt"
        assert node.path == path / "path"

        assert node.file.read_text() == "Hello, World!"
        assert (node.path / "data.txt").read_text() == "Lorem Ipsum"



"""Creating a Node with invalid parameters should raise an error."""

import pathlib

import pandas as pd
import pytest

import zntrack


class MyNode(zntrack.Node):
    # required
    parameter: int = zntrack.params()
    parameter_path: str = zntrack.params_path()

    deps_path: str = zntrack.deps_path()
    outs_path: str = zntrack.outs_path()
    metrics_path: str = zntrack.metrics_path()
    plots_path: str = zntrack.plots_path()

    # optional
    opt_parameter: int = zntrack.params(1)
    opt_parameter_path: str = zntrack.params_path("parameter.yaml")

    opt_deps_path: str = zntrack.deps_path("deps.yaml")
    opt_outs_path: str = zntrack.outs_path("outs.yaml")
    opt_metrics_path: str = zntrack.metrics_path("my_metrics.json")
    opt_plots_path: str = zntrack.plots_path("my_plots.csv")

    # not allowed
    outs: int = zntrack.outs()
    metrics: dict = zntrack.metrics()
    plots: pd.DataFrame = zntrack.plots(y="y")


class SimpleMyNode(zntrack.Node):
    outs_path_a: str = zntrack.outs_path()
    outs_path_b: str = zntrack.outs_path()

    def run(self):
        pathlib.Path(self.outs_path_a).write_text("a")
        pathlib.Path(self.outs_path_b).write_text("b")


def test_init():
    with pytest.raises(TypeError):
        MyNode()  # missing required parameters

    # works
    MyNode(
        parameter=1,
        parameter_path="parameter.yaml",
        deps_path="deps.yaml",
        outs_path="outs.yaml",
        metrics_path="my_metrics.json",
        plots_path="my_plots.csv",
    )
    # works with optional
    MyNode(
        parameter=1,
        parameter_path="parameter.yaml",
        deps_path="deps.yaml",
        outs_path="outs.yaml",
        metrics_path="my_metrics.json",
        plots_path="my_plots.csv",
        opt_parameter=1,
        opt_parameter_path="parameter_1.yaml",
        opt_deps_path="deps_1.yaml",
        opt_outs_path="outs_1.yaml",
        opt_metrics_path="my_metrics_1.json",
        opt_plots_path="my_plots_1.csv",
    )

    # fails with not allowed
    with pytest.raises(TypeError):
        MyNode(
            parameter=1,
            parameter_path="parameter.yaml",
            deps_path="deps.yaml",
            outs_path="outs.yaml",
            metrics_path="my_metrics.json",
            plots_path="my_plots.csv",
            # not allowed
            outs=1,
        )
    with pytest.raises(TypeError):
        MyNode(
            parameter=1,
            parameter_path="parameter.yaml",
            deps_path="deps.yaml",
            outs_path="outs.yaml",
            metrics_path="my_metrics.json",
            plots_path="my_plots.csv",
            # not allowed
            metrics={},
        )
    with pytest.raises(TypeError):
        MyNode(
            parameter=1,
            parameter_path="parameter.yaml",
            deps_path="deps.yaml",
            outs_path="outs.yaml",
            metrics_path="my_metrics.json",
            plots_path="my_plots.csv",
            # not allowed
            plots=pd.DataFrame(),
        )


@pytest.mark.xfail(
    reason="Duplicate outs are currently filtered, but should raise an error."
)
def test_duplicate_outs_paths(proj_path):
    with pytest.raises(ValueError):
        with zntrack.Project() as proj:
            SimpleMyNode(
                outs_path_a="file.txt",
                outs_path_b="file.txt",
            )
        proj.build()


if __name__ == "__main__":
    test_duplicate_outs_paths("proj_path")



import os

import git
import pytest

import zntrack.examples


@pytest.mark.xfail(reason="pending implementation")
def test_patch_open(proj_path):
    with zntrack.Project() as proj:
        node = zntrack.examples.WriteDVCOuts(params="Hello World")

    proj.run()

    repo = git.Repo()

    repo.git.add(A=True)
    commit = repo.index.commit("initial commit")

    node.params = "Lorem Ipsum"
    proj.run()

    node = zntrack.examples.WriteDVCOuts.from_rev(rev=commit.hexsha)

    with node.state.magic_patch():
        with open(node.outs, "r") as f:
            assert f.read() == "Hello World"

    with open(node.outs, "r") as f:
        assert f.read() == "Lorem Ipsum"

    listdir = os.listdir(node.nwd)

    with node.state.magic_patch():
        os.listdir(node.nwd) == listdir


@pytest.mark.xfail(reason="pending implementation")
def test_patch_list(proj_path):
    node = zntrack.from_rev(
        "HelloWorld",
        remote="https://github.com/PythonFZ/ZnTrackExamples.git",
        rev="890c714",
    )

    def func(self, path):
        return os.listdir(path)

    type(node).list = func

    with node.state.magic_patch():
        assert "nodes/HelloWorld/random_number.json" in node.list(node.nwd)
        assert "nodes/HelloWorld/node-meta.json" in node.list(node.nwd)

    assert node.list(node.nwd) == []



import subprocess

import zntrack.examples


def test_no_restart(tmp_path_2):
    project = zntrack.Project()

    with project:
        node = zntrack.examples.NodeWithRestart(start=0)

    assert node.state.run_count == 0

    project.run()

    # node.load()
    assert node.count == 1
    assert node.state.run_count == 1
    assert node.state.restarted is False


def test_restarts(tmp_path_2):
    project = zntrack.Project()

    with project:
        node = zntrack.examples.NodeWithRestart(start=0, raise_exception_until=2)

    project.build()
    # can not longer use project.run() because it will remove outputs, running manually
    # this should raise an exception
    results = subprocess.run(
        ["zntrack", "run", "zntrack.examples.NodeWithRestart", "--name=NodeWithRestart"]
    )
    assert results.returncode == 1
    # this should succeed
    results = subprocess.run(
        ["zntrack", "run", "zntrack.examples.NodeWithRestart", "--name=NodeWithRestart"]
    )
    assert results.returncode == 0

    node = node.from_rev(node.name)
    assert node.count == 2
    assert node.state.run_count == 2
    assert node.state.restarted is True



import dataclasses

import yaml

from zntrack.config import PARAMS_FILE_PATH, FieldTypes
from zntrack.fields.base import field
from zntrack.node import Node


def _params_getter(self: "Node", name: str):
    with self.state.fs.open(PARAMS_FILE_PATH) as f:
        return yaml.safe_load(f)[self.name][name]


def params(default=dataclasses.MISSING, **kwargs):
    """ZnTrack parameter field.

    A field to define a parameter for a ZnTrack node.

    Parameters
    ----------
    default : dict|int|str|float|list|None, optional
        Set a default parameter value.

    default_factory : callable, optional
        A callable that returns the default value.
        Should be used instead of `default` if the default value is mutable.

    Examples
    --------

    >>> import zntrack
    >>> class MyNode(zntrack.Node):
    ...     param: int = zntrack.params(default=42)
    ...
    ...     def run(self) -> None: ...
    ...
    >>> a = MyNode()
    >>> a.param
    42
    >>> b = MyNode(param=43)
    >>> b.param
    43

    """
    # TODO: check types, do not allow e.g. connections
    #  or anything that can not be serialized
    return field(
        default=default,
        field_type=FieldTypes.PARAMS,
        load_fn=_params_getter,
        suffix=None,
        cache=None,
        independent=None,
        **kwargs,
    )



"""The ZnTrack CLI."""

import contextlib
import datetime
import importlib.metadata
import os
import pathlib
import sys

import git
import typer
import yaml

from zntrack import Node, utils
from zntrack.state import PLUGIN_LIST
from zntrack.utils.import_handler import import_handler
from zntrack.utils.misc import load_env_vars

load_env_vars()

app = typer.Typer()


def version_callback(value: bool) -> None:
    """Get the installed 'ZnTrack' version."""
    if value:
        path = pathlib.Path(__file__).parent.parent.parent
        report = f"ZnTrack {importlib.metadata.version('zntrack')} at '{path}'"

        with contextlib.suppress(git.exc.InvalidGitRepositoryError):
            repo = git.Repo(path)
            _ = repo.git_dir

            report += " - "
            with contextlib.suppress(TypeError):  # detached head
                report += f"{repo.active_branch.name}@"
            report += f"{repo.head.object.hexsha[:7]}"
            if repo.is_dirty():
                report += " (dirty)"

        typer.echo(report)
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    ),
) -> None:
    """ZnTrack CLI main callback."""
    _ = version  # this would be greyed out otherwise


@app.command()
def run(
    node_path: str, name: str = None, meta_only: bool = False, method: str = "run"
) -> None:
    """Execute a ZnTrack Node.

    Use as 'zntrack run module.Node --name node_name'.

    Arguments:
    ---------
    node_path : str
        The full path to the Node, e.g. `ipsuite.nodes.SmilesToAtoms`
    name : str
        The name of the node.
    meta_only : bool
        Save only the metadata.
    method : str, default 'run'
        The method to run on the node.

    """
    start_time = datetime.datetime.now()
    utils.misc.load_env_vars(name)
    sys.path.append(pathlib.Path.cwd().as_posix())

    cls: Node = utils.import_handler.import_handler(node_path)
    node: Node = cls.from_rev(name=name, running=True)
    node.state.increment_run_count()
    node.state.save_node_meta()
    # dynamic version of node.run()
    getattr(node, method)()
    run_time = datetime.datetime.now() - start_time
    node.state.add_run_time(run_time)
    node.state.save_node_meta()
    node.save()


@app.command()
def list(
    remote: str = typer.Argument(".", help="The path/url to the repository"),
    rev: str = typer.Argument(None, help="The revision to list (default: HEAD)"),
):
    """List all Nodes in the Project."""
    groups, _ = utils.cli.get_groups(remote, rev)
    print(yaml.dump(groups))


@app.command()
def finalize(
    skip_cached: bool = typer.Option(True, help="Do not upload cached nodes."),
    update_run_names: bool = typer.Option(
        True, help="Include part of the commit message in the run names."
    ),
):
    """Post-commit step for plugin integration."""
    utils.misc.load_env_vars()
    plugins_paths = os.environ.get(
        "ZNTRACK_PLUGINS", "zntrack.plugins.dvc_plugin.DVCPlugin"
    )
    plugins: PLUGIN_LIST = [import_handler(p) for p in plugins_paths.split(",")]
    for plugin in plugins:
        plugin.finalize(skip_cached=skip_cached, update_run_names=update_run_names)



import uuid

import dvc.scm
import pytest

import zntrack.examples

# from zntrack.utils import NodeStatusResults


@pytest.mark.needs_internet
def test_module_not_installed():
    with pytest.raises(ModuleNotFoundError):
        zntrack.from_rev(
            "ASEMD",
            remote="https://github.com/IPSProjects/IPS-Water",
            rev="ca0eef0ccfcbfb72a82136849a9ca35eac8b7629",
        )


@pytest.mark.needs_internet
def test_commit_not_found():
    with pytest.raises(dvc.scm.RevError):
        zntrack.from_rev(
            "ASEMD",
            remote="https://github.com/IPSProjects/IPS-Water",
            rev="this-does-not-exist",
        )


def test_import_from_remote(proj_path):
    node: zntrack.examples.ParamsToMetrics = zntrack.from_rev(
        "ParamsToMetrics",
        remote="https://github.com/PythonFZ/zntrack-examples",
        rev="8d0c992",
    )
    assert node.params == {"loss": 0.1, "accuracy": 0.9}
    assert node.metrics == {"loss": 0.1, "accuracy": 0.9}
    assert node.name == "ParamsToMetrics"
    assert node.state.rev == "8d0c992"
    assert node.state.remote == "https://github.com/PythonFZ/zntrack-examples"
    assert node.uuid == uuid.UUID("65b1c652-6508-4ee5-816c-c2f3cec22cc7")
    assert (
        node.state.get_stage_hash()
        == "70cbf7993d07a6cd0266a0fc0a874e163bc6f464ecc82bb1367b18d24091853c"
    )
    assert (
        node.state.get_stage_hash(include_outs=True)
        == "0e2ec8fab1123c1259ccf96a9590c4b161fc44cf4d93f755699a8fe99c3afe4c"
    )
    # assert node.state.state ==


def test_connect_from_remote(proj_path):
    project = zntrack.Project()

    external_node: zntrack.examples.ParamsToMetrics = zntrack.from_rev(
        name="ParamsToMetrics",
        remote="https://github.com/PythonFZ/zntrack-examples",
        rev="59dac86",
    )

    with project:
        n1 = zntrack.examples.DepsToMetrics(
            deps=external_node.metrics,
        )
        n2 = zntrack.examples.DepsToMetrics(
            deps=external_node.metrics,
        )

    project.repro()

    assert external_node.metrics == {"accuracy": 0.9, "loss": 0.1}
    assert n1.metrics == {"accuracy": 0.9, "loss": 0.1}
    assert n2.metrics == {"accuracy": 0.9, "loss": 0.1}

    # TODO: test with different rev and assert that things changed!

    # raise NotImplementedError("This test is not done yet!")

    # node_a = zntrack.from_rev(
    #     "HelloWorld",
    #     remote="https://github.com/PythonFZ/ZnTrackExamples.git",
    #     rev="890c714",
    # )

    # node_b = zntrack.from_rev(
    #     "HelloWorld",
    #     remote="https://github.com/PythonFZ/ZnTrackExamples.git",
    #     rev="369fe8f",
    # )

    # assert node_a.random_number == 123
    # assert node_b.random_number == 126

    # with zntrack.Project() as project:
    #     node = zntrack.examples.AddOne(number=node_a.random_number)

    # project.repro()

    # assert node.outs == node_a.random_number + 1

    # with zntrack.Project() as project:
    #     node = zntrack.examples.AddOne(number=node_b.random_number)

    # project.run()
    # # We can not use node.load() here and build again,
    # # because it will convert connections to e.g. type int
    # # and then we can not connect to the node anymore.
    # # node.load()

    # assert zntrack.from_rev(node.name).outs == node_b.random_number + 1

    # project.build()


def test_two_nodes_connect_external(proj_path):
    node_a: zntrack.examples.ParamsToOuts = zntrack.from_rev(
        name="NumericOuts",
        remote="https://github.com/PythonFZ/zntrack-examples",
        rev="de82dc7104ac3",
    )

    with zntrack.Project() as project:
        node1 = zntrack.examples.AddOne(number=node_a.outs)
        node2 = zntrack.examples.AddOne(number=node_a.outs)

    project.repro()

    assert node1.outs == node_a.outs + 1
    assert node2.outs == node_a.outs + 1



import contextlib
import dataclasses
import warnings
from dataclasses import Field, dataclass
from typing import Any

import dvc.repo
import git
import mlflow
import pandas as pd
import znflow
from mlflow.utils import mlflow_tags

from zntrack.config import (
    FIELD_TYPE,
    PLUGIN_EMPTY_RETRUN_VALUE,
    FieldTypes,
)
from zntrack.node import Node
from zntrack.plugins import ZnTrackPlugin, get_exp_info, set_exp_info
from zntrack.utils import module_handler
from zntrack.utils.misc import load_env_vars

# TODO: if this plugin fails, there should only be a warning, not an error
# so that the results are not lost
# TODO: have the mlflow run active over the entire run
#  method to avoid searching for it over again.
# TODO: in finalize have the parent run active (if not already)


def get_mlflow_parent_run() -> str:
    # Can I get rid of this by using cwd and machine_id?

    exp_info = get_exp_info()
    try:
        parent_run_id = exp_info["parent_run_id"]
        mlflow.start_run(run_id=parent_run_id)
        return parent_run_id

    except KeyError:
        run = mlflow.start_run()
        parent_run_id = run.info.run_id
        exp_info["parent_run_id"] = parent_run_id
        set_exp_info(exp_info)

        # set tags
        tags = exp_info.get("tags", {})
        for tag_key, tag_value in tags.items():
            mlflow.set_tag(tag_key, tag_value)
        return parent_run_id


def get_mlflow_child_run(stage_hash: str, node: Node, node_path: str) -> str:
    runs_df = mlflow.search_runs(
        filter_string=f"tags.dvc_stage_hash = '{stage_hash}'",
    )
    if len(runs_df) == 0:
        # we assume there is an active run, maybe test?
        run = mlflow.start_run(nested=True)
        warnings.warn("Creating new child run")
        exp_info = get_exp_info()
        tags = exp_info.get("tags", {})

        mlflow.set_tag("dvc_stage_hash", stage_hash)
        mlflow.set_tag("dvc_stage_name", node.name)
        # TODO: do we want to include the name of the parent run?
        mlflow.set_tag(mlflow_tags.MLFLOW_RUN_NAME, node.name)
        mlflow.set_tag(
            "zntrack_node",
            node_path,
        )
        for tag_key, tag_value in tags.items():
            mlflow.set_tag(tag_key, tag_value)
        return run.info.run_id
    else:
        print("found existing run")
        run_id = runs_df.iloc[0].run_id
        mlflow.start_run(run_id=run_id, nested=True)
        return run_id


@dataclass
class MLFlowPlugin(ZnTrackPlugin):
    """ZnTrack integration with MLFlow.

    ```mermaid
    flowchart LR
        repo --> commit --> stage

        subgraph Experiment
        repo
        end
        subgraph Parent Run
        commit
        end
        subgraph Child Run
        stage
        end
    ```
    """

    _continue_on_error_ = True
    parent_run_id = None
    child_run_id = None

    def __post_init__(self):
        load_env_vars("")
        # can not load the runs in here, because the node name is not set yet.

    def setup(self):
        # 1. find the parent run, create one if it does not exist
        # 2. find the child run, create one if it does not exist

        self.parent_run_id = self.parent_run_id or get_mlflow_parent_run()
        self.child_run_id = self.child_run_id or get_mlflow_child_run(
            self.node.state.get_stage_hash(),
            self.node,
            f"{self.node.__module__}.{self.node.__class__.__name__}",
        )

    def close(self):
        # close both the parent and the child run
        while mlflow.active_run():
            mlflow.end_run()

    def get_run_info(self) -> dict:
        # get the name of the current run
        run_info = mlflow.active_run().info
        experiment_id = run_info.experiment_id
        # run_name = run_info.run_name
        try:
            run_name = mlflow.get_run(run_info.run_id).data.tags[
                mlflow_tags.MLFLOW_RUN_NAME
            ]
        except KeyError:
            run_name = run_info.run_name
        run_id = run_info.run_id
        host_url = mlflow.get_tracking_uri()
        uri = f"{host_url}/#/experiments/{experiment_id}/runs/{run_id}"
        return {"name": run_name, "uri": uri, "id": run_id}

    def getter(self, field: Field) -> Any:
        return PLUGIN_EMPTY_RETRUN_VALUE

    def save(self, field: Field) -> None:
        if field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS:
            mlflow.log_param(field.name, getattr(self.node, field.name))
        if field.metadata.get(FIELD_TYPE) == FieldTypes.DEPS:
            content = getattr(self.node, field.name)
            new_content = []
            if not isinstance(content, (list, tuple)):
                content = [content]
                # TODO: code duplication with the other plugins!
            for val in content:
                if dataclasses.is_dataclass(val) and not isinstance(
                    val, (Node, znflow.Connection, znflow.CombinedConnections)
                ):
                    # We save the values of the passed dataclasses
                    #  to the params.yaml file to be later used
                    #  by the DataclassContainer to recreate the
                    #  instance with the correct parameters.
                    dc_params = dataclasses.asdict(val)
                    dc_params["_cls"] = f"{module_handler(val)}.{val.__class__.__name__}"
                    new_content.append(dc_params)
            mlflow.log_param(field.name, new_content)

        if field.metadata.get(FIELD_TYPE) == FieldTypes.METRICS:
            metrics = getattr(self.node, field.name)
            for key, value in metrics.items():
                mlflow.log_metric(f"{field.name}.{key}", value)
                # TODO: plots
                # TODO: define tags for all experiments in a parent run
        if field.metadata.get(FIELD_TYPE) == FieldTypes.PLOTS:
            df: pd.DataFrame = getattr(self.node, field.name).copy()
            for idx, row in df.iterrows():
                for key, value in row.items():
                    mlflow.log_metric(f"{field.name}.{key}", value, step=idx)

    def convert_to_dvc_yaml(self):
        return PLUGIN_EMPTY_RETRUN_VALUE

    def convert_to_params_yaml(self):
        return PLUGIN_EMPTY_RETRUN_VALUE

    def convert_to_zntrack_json(self, graph):
        exp_info = get_exp_info()
        if len(graph.tags) > 0:
            exp_info["tags"] = graph.tags
            set_exp_info(exp_info)

        return PLUGIN_EMPTY_RETRUN_VALUE

    @classmethod
    def finalize(cls, **kwargs):
        """Example:
        -------
        python -c "from zntrack.plugins.mlflow_plugin \
                import MLFlowPlugin; MLFlowPlugin.finalize()"

        """
        # TODO: with the dependency on the file this does not support revs
        # We could add all dvc_stage_hashes to the tags of the parent run
        # and then find the parent run by querying the tags

        # TODO: provide post-commit installation instructions
        # TODO: kwarg for updating cached runs
        load_env_vars()
        skip_cached = kwargs.get("skip_cached", True)
        update_run_names = kwargs.get("update_run_names", True)

        exp_info = get_exp_info()
        if "parent_run_id" not in exp_info:
            raise ValueError("Unable to find parent run id")
        import zntrack

        tags = exp_info.get("tags", {})

        repo = git.Repo(".")
        if repo.is_dirty():
            raise ValueError("Can only finalize a clean git repository")

        commit_hash = repo.head.commit.hexsha
        commit_message = repo.head.commit.message

        prefix = commit_message.split()[0]

        # get the default remote for the current branch
        try:
            remote_url = repo.remote().url
        except ValueError:
            # ValueError: Remote named 'origin' didn't exist
            remote_url = None

        node_names = []
        with dvc.repo.Repo() as repo:
            for stage in repo.index.stages:
                with contextlib.suppress(AttributeError):
                    # only PipelineStages have a name attribute
                    #  we do not need data stages
                    node_names.append(stage.name)

        for node_name in node_names:
            node = zntrack.from_rev(node_name)
            plugin = cls(node)

            with plugin:
                pass  # load run_id states
            with mlflow.start_run(run_id=plugin.parent_run_id):
                mlflow.set_tag("git_commit_hash", commit_hash)
                mlflow.set_tag("git_commit_message", commit_message)
                mlflow.set_tag(mlflow_tags.MLFLOW_RUN_NAME, prefix)
                if remote_url is not None:
                    mlflow.set_tag("git_remote", remote_url)
                parent_run_id = mlflow.active_run().info.run_id
                active_experiment_id = mlflow.active_run().info.experiment_id
            break

        child_runs = mlflow.search_runs(
            experiment_ids=[active_experiment_id],
            filter_string=f"tags.{mlflow_tags.MLFLOW_PARENT_RUN_ID} = '{parent_run_id}'",
        )
        print(f"found {len(child_runs)} child runs in exp: '{active_experiment_id}'")
        for node_name in node_names:
            if node_name in child_runs["tags.dvc_stage_name"].values:
                node = zntrack.from_rev(node_name, rev=commit_hash)
                with node.state.plugins["MLFlowPlugin"]:
                    mlflow.set_tag("git_commit_hash", commit_hash)
                    mlflow.set_tag("git_commit_message", commit_message.strip())
                    # mlflow.set_tag(
                    #     mlflow_tags.MLFLOW_RUN_NOTE, f"{commit_message.strip()}"
                    # )
                    # there is https://github.com/mlflow/mlflow/blob/master/tests/tracking/context/test_git_context.py
                    #  but it is not documented?
                    mlflow.set_tag(mlflow_tags.MLFLOW_GIT_COMMIT, commit_hash)
                    if update_run_names:
                        mlflow.set_tag(
                            mlflow_tags.MLFLOW_RUN_NAME, f"{prefix}:{node_name}"
                        )
                    if remote_url is not None:
                        mlflow.set_tag("git_remote", remote_url)
                        mlflow.set_tag(mlflow_tags.MLFLOW_GIT_REPO_URL, remote_url)
                    if hasattr(node, "__run_note__"):
                        mlflow.set_tag(mlflow_tags.MLFLOW_RUN_NOTE, node.__run_note__())
                    # if hasattr(node, "_mlflow_run_note_"):
                    # raise ValueError("not implemented")
            elif not skip_cached:
                print(f"missing {node_name}")
                node = zntrack.from_rev(node_name, rev=commit_hash)
                stage_hash = node.state.get_stage_hash()
                original_runs = mlflow.search_runs(
                    experiment_ids=[active_experiment_id],
                    filter_string=f"tags.dvc_stage_hash = '{stage_hash}'",
                )
                print(f"searching for original run {stage_hash=}")
                if len(original_runs) == 0:
                    warnings.warn(f"no original run found for {node_name} - skipping")
                    continue
                if len(original_runs) > 1:
                    warnings.warn(
                        f"multiple original runs found for {node_name} - skipping"
                    )
                    continue
                print(f"found original run for {node_name} - {original_runs}")
                original_run_id = original_runs.iloc[0].run_id

                get_mlflow_parent_run()
                with mlflow.start_run(nested=True):
                    mlflow.set_tag("dvc_stage_name", node.name)
                    mlflow.set_tag(
                        "zntrack_node",
                        f"{node.__module__}.{node.__class__.__name__}",
                    )
                    mlflow.set_tag("original_run_id", original_run_id)
                    # copy parameters and metrics from the original run
                    original_run = mlflow.get_run(original_run_id)
                    for key, value in original_run.data.params.items():
                        mlflow.log_param(key, value)
                    for key, value in original_run.data.metrics.items():
                        mlflow.log_metric(key, value)

                    # set git hash and commit message
                    mlflow.set_tag("git_commit_hash", commit_hash)
                    mlflow.set_tag("git_commit_message", commit_message.strip())
                    if remote_url is not None:
                        mlflow.set_tag("git_remote", remote_url)

                    mlflow.set_tag("original_dvc_stage_hash", node.state.get_stage_hash())
                    if hasattr(node, "__run_note__"):
                        mlflow.set_tag(mlflow_tags.MLFLOW_RUN_NOTE, node.__run_note__())
                    mlflow.set_tag(mlflow_tags.MLFLOW_RUN_NAME, node_name)
                    for tag_key, tag_value in tags.items():
                        mlflow.set_tag(tag_key, tag_value)
                while mlflow.active_run():
                    mlflow.end_run()

        # remove the parent run id exp_info
        exp_info.pop("parent_run_id")
        set_exp_info(exp_info)



import json
import pathlib

import numpy as np
import yaml

import zntrack
from zntrack.config import NOT_AVAILABLE, FieldTypes
from zntrack.fields import field

CWD = pathlib.Path(__file__).parent.resolve()


def _h5data_getter(self: zntrack.Node, name: str, suffix: str):
    raise NotImplementedError


def _h5data_save_func(self: zntrack.Node, name: str, suffix: str):
    raise NotImplementedError


def h5data(*, cache: bool = True, independent: bool = False, **kwargs):
    return field(
        cache=cache,
        independent=independent,
        field_type=FieldTypes.OUTS,
        load_fn=_h5data_getter,
        dump_fn=_h5data_save_func,
        suffix=".h5",
        repr=False,
        default=NOT_AVAILABLE,
        **kwargs,
    )


class TextNode(zntrack.Node):
    content: np.ndarray = h5data()

    def run(self):
        self.content = np.arange(9).reshape(3, 3)


def test_text_node(proj_path):
    with zntrack.Project() as project:
        TextNode()
    project.build()

    assert json.loads(
        (CWD / "zntrack_config" / "custom_field.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "custom_field.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "custom_field.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()



import json
import subprocess

import zntrack


class MetricsNode(zntrack.Node):
    n: int = zntrack.params()

    metrics: dict = zntrack.metrics()

    def run(self):
        self.metrics = {"n": self.n}


def test_simple_metrics(proj_path):
    with zntrack.Project() as proj:
        node = MetricsNode(n=10)

    proj.build()
    subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

    with open(node.nwd / "metrics.json") as f:
        metrics = json.load(f)

    assert isinstance(metrics, dict)
    assert metrics["n"] == 10
    assert len(metrics) == 1

    assert isinstance(node.metrics, dict)
    assert node.metrics["n"] == 10
    assert len(node.metrics) == 1

    # create a new instance (lazy loading)
    node = node.from_rev()
    assert isinstance(node.metrics, dict)
    assert node.metrics["n"] == 10
    assert len(node.metrics) == 1



import pathlib
import typing as t

import pytest
import yaml

import zntrack.examples
from zntrack.config import NodeStatusEnum


@pytest.mark.parametrize("eager", [True, False])
def test_AddNumbers(proj_path, eager):
    with zntrack.Project() as project:
        add_numbers = zntrack.examples.AddNumbers(a=1, b=2)

    assert add_numbers.state.state == NodeStatusEnum.CREATED

    if eager:
        project.run()
    else:
        project.repro()

    assert add_numbers.c == 3
    add_numbers = add_numbers.from_rev()
    assert add_numbers.c == 3
    assert add_numbers.state.state == NodeStatusEnum.FINISHED


def test_AddNumbers_remove_params(proj_path):
    with zntrack.Project() as project:
        add_numbers = zntrack.examples.AddNumbers(a=1, b=2)

    assert add_numbers.state.state == NodeStatusEnum.CREATED

    project.repro()

    params = pathlib.Path("params.yaml").read_text()
    params = yaml.safe_load(params)
    params[add_numbers.name] = {}
    params = pathlib.Path("params.yaml").write_text(yaml.dump(params))

    # with pytest.raises(zntrack.exceptions.NodeNotAvailableError):
    #     add_numbers.load()


def test_zntrack_from_rev(proj_path):
    with zntrack.Project() as project:
        add_numbers = zntrack.examples.AddNumbers(a=1, b=2)

    assert add_numbers.state.state == NodeStatusEnum.CREATED

    project.repro()

    node = zntrack.from_rev(add_numbers.name)
    assert node.a == 1
    assert node.b == 2
    assert node.c == 3
    assert node.state.state == NodeStatusEnum.FINISHED


@pytest.mark.parametrize("eager", [True, False])
def test_AddNumbers_named(proj_path, eager):
    with zntrack.Project() as project:
        add_numbers_a = zntrack.examples.AddNumbers(a=1, b=2, name="NodeA")
        add_numbers_b = zntrack.examples.AddNumbers(a=1, b=2, name="NodeB")

    assert add_numbers_a.state.state == NodeStatusEnum.CREATED
    assert add_numbers_b.state.state == NodeStatusEnum.CREATED

    if eager:
        project.run()
    else:
        project.repro()

    assert add_numbers_a.c == 3
    assert add_numbers_b.c == 3


class NodeWithPostInit(zntrack.Node):
    params: t.Any = zntrack.params()
    outs: t.Any = zntrack.outs()

    def __post_init__(self):
        # _ = self.params
        # the value is not loaded in the _post_init_
        self.value = 42

    def run(self) -> None:
        assert self.value == 42
        self.outs = self.params


@pytest.mark.parametrize("eager", [True, False])
def test_NodeWithPostInit(proj_path, eager):
    with zntrack.Project() as project:
        node = NodeWithPostInit(params=10)

    if eager:
        project.run()
    else:
        project.repro()
    node = NodeWithPostInit.from_rev()
    assert node.value == 42
    assert node.outs == 10


class NonDefaultInit(zntrack.Node):
    params: t.Any = zntrack.params()
    outs: t.Any = zntrack.outs()

    # def __init__(self, params, **kwargs):
    #     super().__init__(**kwargs)
    #     self.params = params
    #     self.value = 42

    def run(self) -> None:
        self.outs = self.params


def test_NonDefaultInit(proj_path):
    with zntrack.Project() as project:
        node = NonDefaultInit(params=10)
    project.run()

    node = NonDefaultInit.from_rev()
    with pytest.raises(AttributeError):
        """We can not load an __init__ with non-default parameters"""
        assert node.value == 42
    assert node.outs == 10


class NoOutsWritten(zntrack.Node):
    params: t.Any = zntrack.params()
    outs: t.Any = zntrack.outs()

    def run(self) -> None:
        pass


# @pytest.mark.parametrize("eager", [True, False])
# def test_NoOutsWritten(proj_path, eager):
#     with zntrack.Project() as project:
#         node = NoOutsWritten(params=10)
#     project.run(eager=eager)
#     if eager:
#         node.save()
#     else:
#         node.load()

#     node = NoOutsWritten.from_rev()
#     assert node.outs is None


def test_outs_in_init(proj_path):
    with pytest.raises(TypeError):
        # outs can not be set
        _ = zntrack.examples.AddNumbers(a=1, b=2, outs=3)
    with zntrack.Project():
        with pytest.raises(TypeError):
            # outs can not be set
            _ = zntrack.examples.AddNumbers(a=1, b=2, c=3)  # c is an output



"""Test params such as always_changed, ..."""

import json
import pathlib

import yaml

import zntrack

CWD = pathlib.Path(__file__).parent.resolve()


def test_node(proj_path):
    assert zntrack.config.ALWAYS_CACHE is False
    zntrack.config.ALWAYS_CACHE = True
    assert zntrack.config.ALWAYS_CACHE is True

    # We define the node here, because the config has to be set
    #  before calling zntrack.metrics()
    class MyNode(zntrack.Node):
        """Some Node."""

        metric: dict = zntrack.metrics()
        metrics_path: pathlib.Path = zntrack.metrics_path(zntrack.nwd / "results.json")

        def run(self) -> None:
            self.metric = {"a": 1, "b": 2}

    with zntrack.Project() as proj:
        MyNode()

    proj.build()

    zntrack.config.ALWAYS_CACHE = False  # reset to default value
    assert zntrack.config.ALWAYS_CACHE is False

    assert json.loads(
        (CWD / "zntrack_config" / "user_config.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "user_config.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "user_config.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_node("")



"""Helpers for the Node Working Directory (NWD)."""

import json
import logging
import os
import pathlib
import shutil
import typing as t

import znflow.utils
import znjson

from zntrack.add import DVCImportPath
from zntrack.config import NWD_PATH, ZNTRACK_FILE_PATH, NodeStatusEnum

if t.TYPE_CHECKING:
    from zntrack import Node


log = logging.getLogger(__name__)


class _NWD(str):
    """String that allows for pathlib like truediv operation."""

    def __truediv__(self, other) -> pathlib.Path:
        return pathlib.Path(self, other)


nwd = _NWD("$nwd$")


def move_nwd(target: pathlib.Path, destination: pathlib.Path) -> None:
    """Move files from 'target' to 'destination'."""
    if destination.exists():
        shutil.rmtree(destination)
    shutil.copytree(target, destination, copy_function=os.link)
    shutil.rmtree(target)


def get_nwd(node: "Node") -> pathlib.Path:
    """Get the node working directory.

    This is used instead of `node.nwd` because it allows
    for parameters to define if the nwd should be created.

    Arguments:
    ---------
    node: Node
        The node instance for which the nwd should be returned.

    """
    try:
        nwd = node.__dict__["nwd"]
    except KeyError:
        if node.name is None:
            raise ValueError("Unable to determine node name.")
        if (
            node.state.remote is None
            and node.state.rev is None
            and node.state.state == NodeStatusEnum.FINISHED
        ):
            nwd = pathlib.Path(NWD_PATH, node.name)
        else:
            try:
                with node.state.fs.open(ZNTRACK_FILE_PATH) as f:
                    zntrack_config = json.load(f)
                nwd = zntrack_config[node.name]["nwd"]
                nwd = json.loads(json.dumps(nwd), cls=znjson.ZnDecoder)
            except (FileNotFoundError, KeyError):
                nwd = pathlib.Path(NWD_PATH, node.name)

    if node.state.group is not None:
        # strip the groups from node_name
        to_replace = "_".join(node.state.group.names) + "_"
        replacement = "/".join(node.state.group.names) + "/"
        nwd = pathlib.Path(str(nwd).replace(to_replace, replacement))

    return nwd


class NWDReplaceHandler(znflow.utils.IterableHandler):
    """Replace the nwd placeholder with the actual nwd."""

    def default(self, value, **kwargs):
        """Replace the nwd placeholder with the actual nwd."""
        if isinstance(value, str):
            if value == nwd:
                # nwd is of type str but will be converted to pathlib.Path
                return pathlib.Path(kwargs["nwd"])
            return value.replace(nwd, pathlib.Path(kwargs["nwd"]).as_posix())
        elif isinstance(value, pathlib.Path):
            return pathlib.Path(
                value.as_posix().replace(nwd, pathlib.Path(kwargs["nwd"]).as_posix())
            )
        elif value is None:
            return value
        elif isinstance(value, DVCImportPath):
            # with DVCImportPath, we do not expect the nwd placeholder
            return value.path
        else:
            raise ValueError(
                f"replace_nwd_placeholder is not implemented for {type(value)}"
            )



import contextlib
import dataclasses
import os
import typing as t
import uuid

import aim
import dvc.repo
import git
import pandas as pd
import znflow

from zntrack.config import FIELD_TYPE, PLUGIN_EMPTY_RETRUN_VALUE, FieldTypes
from zntrack.node import Node
from zntrack.plugins import ZnTrackPlugin, get_exp_info, set_exp_info
from zntrack.utils import module_handler
from zntrack.utils.misc import load_env_vars

if t.TYPE_CHECKING:
    pass


def _create_aim_run(
    aim_repo, experiment, stage_hash, stage_name, node_path, tags: dict[str, str]
) -> str:
    assert experiment.startswith("exp-")
    run = aim.Run(repo=aim_repo, experiment=experiment)
    run["dvc_stage_hash"] = stage_hash
    run["dvc_stage_name"] = stage_name
    run["zntrack_node"] = node_path
    for tag_key, tag_value in tags.items():
        run.add_tag(f"{tag_key}={tag_value}")
    run.close()
    return run.hash


def get_original_run_data(stage_hash: str) -> pd.DataFrame:
    """Retrieve parameters from the original run."""
    load_env_vars()
    aim_repo = aim.Repo(path=os.environ["AIM_TRACKING_URI"])
    for run_metrics_col in aim_repo.query_metrics(
        f"run.dvc_stage_hash == '{stage_hash}'"
    ).iter():
        return run_metrics_col.run.dataframe()
    raise ValueError(f"Unable to find run with dvc_stage_hash {stage_hash}")


def get_aim_run_id(stage_hash: str, stage_name: str, node_path: str) -> tuple[str, bool]:
    load_env_vars()
    aim_repo = aim.Repo(path=os.environ["AIM_TRACKING_URI"])
    exp_info = get_exp_info()
    experiment = exp_info.get("aim_experiment")
    tags = exp_info.get("tags", {})
    if experiment is None:
        experiment = f"exp-{uuid.uuid4().hex[:8]}"
        exp_info["aim_experiment"] = experiment
        set_exp_info(exp_info)
        return (
            _create_aim_run(
                aim_repo, experiment, stage_hash, stage_name, node_path, tags
            ),
            True,
        )
    else:
        for run_metrics_col in aim_repo.query_metrics(
            f"run.dvc_stage_hash == '{stage_hash}' and run.experiment == '{experiment}'"
        ).iter():
            return run_metrics_col.run.hash, False
        return (
            _create_aim_run(
                aim_repo, experiment, stage_hash, stage_name, node_path, tags
            ),
            True,
        )


class AIMPlugin(ZnTrackPlugin):
    """ZnTrack integration with AIM.

    ```mermaid
    flowchart LR
        repo --> commit --> stage

        subgraph AIM repo
        repo
        end
        subgraph Experiment
        commit
        end
        subgraph Run
        stage
        end
    ```
    """

    _continue_on_error_ = True

    def setup(self):
        self.run = None
        self.run_id, self.new_run = get_aim_run_id(
            self.node.state.get_stage_hash(),
            self.node.name,
            f"{self.node.__module__}.{self.node.__class__.__name__}",
        )

    def getter(self, field: dataclasses.Field) -> t.Any:
        # The AIM plugin relies on others, e.g. DVCPlugin for loading/saving data
        return PLUGIN_EMPTY_RETRUN_VALUE

    def save(self, field: dataclasses.Field) -> None:
        if field.metadata.get(FIELD_TYPE) == FieldTypes.PARAMS:
            with self.get_aim_run() as run:
                run[field.name] = getattr(self.node, field.name)
        if field.metadata.get(FIELD_TYPE) == FieldTypes.DEPS:
            content = getattr(self.node, field.name)
            new_content = []
            if not isinstance(content, (list, tuple)):
                content = [content]
                # TODO: code duplication with the other plugins!
            for val in content:
                if dataclasses.is_dataclass(val) and not isinstance(
                    val, (Node, znflow.Connection, znflow.CombinedConnections)
                ):
                    # We save the values of the passed dataclasses
                    #  to the params.yaml file to be later used
                    #  by the DataclassContainer to recreate the
                    #  instance with the correct parameters.
                    dc_params = dataclasses.asdict(val)
                    dc_params["_cls"] = f"{module_handler(val)}.{val.__class__.__name__}"
                    new_content.append(dc_params)
            with self.get_aim_run() as run:
                run[field.name] = (
                    new_content
                    if isinstance(getattr(self.node, field.name), list)
                    else new_content[0]
                )
        if field.metadata.get(FIELD_TYPE) == FieldTypes.METRICS:
            with self.get_aim_run() as run:
                assert isinstance(getattr(self.node, field.name), dict)

                for key, value in getattr(self.node, field.name).items():
                    run.track(value, name=f"{field.name}.{key}")

        if field.metadata.get(FIELD_TYPE) == FieldTypes.PLOTS:
            df: pd.DataFrame = getattr(self.node, field.name).copy()
            with self.get_aim_run() as run:
                for idx, row in df.iterrows():
                    for key, value in row.items():
                        run.track(value, name=f"{field.name}.{key}", step=idx)

    @contextlib.contextmanager
    def get_aim_run(self) -> t.Iterator[aim.Run]:
        if not hasattr(self, "run_id") or self.run_id is None:
            self.setup()
        exp_info = get_exp_info()
        experiment = exp_info.get("aim_experiment")
        run = aim.Run(
            run_hash=self.run_id,
            repo=aim.Repo(path=os.environ["AIM_TRACKING_URI"]),
            experiment=experiment,
        )
        yield run
        run.close()

    def convert_to_dvc_yaml(self):
        return PLUGIN_EMPTY_RETRUN_VALUE

    def convert_to_params_yaml(self):
        return PLUGIN_EMPTY_RETRUN_VALUE

    def convert_to_zntrack_json(self, graph):
        exp_info = get_exp_info()
        if len(graph.tags) > 0:
            exp_info["tags"] = graph.tags
            set_exp_info(exp_info)

        return PLUGIN_EMPTY_RETRUN_VALUE

    @classmethod
    def finalize(cls, **kwargs) -> None:
        """Update the aim run to include the correct commit data.

        Run this, once you have created a commit for your DVC experiment.
        This will update the aim run with the correct git hash, message and remote.

        Parameters
        ----------
        rev : str, optional
            The git revision to use. If None, 'HEAD' will be used.

        Example:
        -------
        python -c "from zntrack.plugins.aim_plugin import AIMPlugin; AIMPlugin.finalize()"

        """
        load_env_vars()

        exp_info = get_exp_info()
        if "aim_experiment" not in exp_info:
            raise ValueError("Unable to find an AIM experiment.")

        import zntrack

        exp_info.get("tags", {})

        repo = git.Repo(".")
        if repo.is_dirty():
            raise ValueError("Can only finalize a clean git repository")

        commit_hash = repo.head.commit.hexsha
        commit_message = repo.head.commit.message

        # get the default remote for the current branch
        try:
            remote_url = repo.remote().url
        except ValueError:
            # ValueError: Remote named 'origin' didn't exist
            remote_url = None

        node_names = []
        with dvc.repo.Repo() as repo:
            for stage in repo.index.stages:
                with contextlib.suppress(AttributeError):
                    # only PipelineStages have a name attribute
                    #  we do not need data stages
                    node_names.append(stage.name)
        for node_name in node_names:
            # TODO: this will create all nodes, even those, that were cached.
            # currently, they are missing parameters and original_run_id
            # but they will have the dvc_stage_hash
            # should be original_dvc_stage_hash and original_run_id instead.
            # and the parameters should be filled in from the original run

            node = zntrack.from_rev(node_name)
            plugin = cls(node)

            with plugin.get_aim_run() as run:
                if plugin.new_run:
                    # move dvc_stage_hash to original_dvc_stage_hash
                    run["original_dvc_stage_hash"] = run["dvc_stage_hash"]
                    del run["dvc_stage_hash"]
                    original_run_df = get_original_run_data(node.state.get_stage_hash())
                    run_df = run.dataframe()
                    for key, value in original_run_df.items():
                        if key in ["dvc_stage_name", "dvc_stage_hash", "zntrack_node"]:
                            continue
                        if key not in run_df.columns:
                            run[key] = value[0]
                    run["original_run_id"] = original_run_df["hash"][0]
                run["git_commit_hash"] = commit_hash
                run["git_commit_message"] = commit_message.strip()
                if remote_url is not None:
                    run["git_remote"] = remote_url

        exp_info.pop("aim_experiment")
        set_exp_info(exp_info)

        # import zntrack

        # repo = git.Repo(".")

        # if rev is None:
        #     commit_hash = repo.head.commit.hexsha
        #     commit_message = repo.head.commit.message
        # else:
        #     commit_hash = repo.commit(rev).hexsha
        #     commit_message = repo.commit(rev).message

        # # get the default remote for the current branch
        # try:
        #     remote_url = repo.remote().url
        # except ValueError:
        #     # ValueError: Remote named 'origin' didn't exist
        #     remote_url = None

        # node_names = []
        # with dvc.repo.Repo(rev=rev) as repo:
        #     for stage in repo.index.stages:
        #         with contextlib.suppress(AttributeError):
        #             # only PipelineStages have a name attribute
        #             # we do not need e.g. data stages
        #             node_names.append(stage.name)
        # for node_name in node_names:
        #     node = zntrack.from_rev(node_name, rev=rev)
        #     plugin = cls(node)
        #     with plugin.get_aim_run() as run:
        #         run["git_commit_hash"] = commit_hash
        #         run["git_commit_message"] = commit_message.strip()
        #         if remote_url is not None:
        #             run["git_remote"] = remote_url
        #         run.experiment = commit_hash
        #     # raise ValueError(f"Updating {node.name} commit hash to {commit_hash}")
        #     # what if the experiment is already set?
        #     # update or keep?

        # # TODO: remove exp key from exp_info



import json
import pathlib

import yaml

import zntrack
import zntrack.examples
import zntrack.group

CWD = pathlib.Path(__file__).parent.resolve()


def test_apply(proj_path) -> None:
    project = zntrack.Project()

    JoinedParamsToOuts = zntrack.apply(zntrack.examples.ParamsToOuts, "join")  # noqa N806

    with project:
        zntrack.examples.ParamsToOuts(params=["a", "b"])
        JoinedParamsToOuts(params=["a", "b"])
        zntrack.apply(zntrack.examples.ParamsToOuts, "join")(params=["a", "b", "c"])

    project.build()

    assert json.loads((CWD / "zntrack_config" / "apply.json").read_text()) == json.loads(
        (proj_path / "zntrack.json").read_text()
    )
    assert yaml.safe_load(
        (CWD / "dvc_config" / "apply.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "apply.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_apply("")



import dataclasses
import shutil
import typing as t

import pytest

import zntrack
from zntrack import Node, Project


@dataclasses.dataclass
class NodeViaParams:
    param1: t.Any = zntrack.params()


class ExampleNode(Node):
    params1: t.Any = zntrack.deps()
    params2: t.Any = zntrack.deps()

    outs: t.Any = zntrack.outs()

    def run(self):
        self.outs = self.params1 + self.params2


class ExampleNodeLst(Node):
    params: list = zntrack.deps()
    outs: int = zntrack.outs()

    def run(self):
        self.outs = sum(p for p in self.params)


@pytest.mark.xfail(reason="pending implementation")
@pytest.mark.parametrize("eager", [True, False])
def test_ExampleNode(proj_path, eager):
    project = Project()
    parameter_1 = NodeViaParams(param1=1)
    parameter_2 = NodeViaParams(param1=10)

    with project:
        node = ExampleNode(params1=parameter_1, params2=parameter_2)

    project.run(eager=eager)
    # if not eager:
    #     node.load()
    assert node.params1.param1 == 1
    assert node.params2.param1 == 10
    assert node.outs == 11

    assert node.params1.name == "ExampleNode+params1"
    assert node.params2.name == "ExampleNode+params2"

    if not eager:
        # Check new instance also works
        node = node.from_rev()
        assert node.params1.param1 == 1
        assert node.params2.param1 == 10
        assert node.outs == 11

        assert node.params1.name == "ExampleNode+params1"
        assert node.params2.name == "ExampleNode+params2"


@pytest.mark.xfail(reason="pending implementation")
@pytest.mark.parametrize("eager", [False])
def test_ExampleNodeLst(proj_path, eager):
    project = Project()
    parameter_1 = NodeViaParams(param1=1)
    parameter_2 = NodeViaParams(param1=10)

    with project:
        node = ExampleNodeLst(params=[parameter_1.param1, parameter_2.param1])

    if eager:
        project.run()
    else:
        project.repro()
    assert node.params[0] == 1
    assert node.params[1] == 10
    assert node.outs == 11

    parameter_1.param1 = 2

    with zntrack.Project() as project:
        node = ExampleNodeLst(params=[parameter_1.param1, parameter_2.param1])
        # passing a parameter of a dataclass is not possible

    if eager:
        project.run()
    else:
        project.repro()
        node = node.from_rev()
    assert node.params[0] == 2
    assert node.params[1] == 10
    assert node.outs == 12

    # assert isinstance(parameter_1, NodeViaParams)
    # with project:
    #     # #     # node = ExampleNodeLst(params=[parameter_1, parameter_2])
    #     node.params = [parameter_1, parameter_2]
    # project.run(eager=eager)

    # # if not eager:
    # #     node.load()
    # assert node.params[0].param1 == 2
    # assert node.params[1].param1 == 10
    # assert node.outs == 12

    # if not eager:
    #     # Check new instance also works
    #     node = node.from_rev()
    #     assert node.params[0].param1 == 2
    #     assert node.params[1].param1 == 10
    #     assert node.outs == 12


@dataclasses.dataclass
class RandomNumberGen:
    seed: int = 42

    def get_rnd(self):
        import random

        random.seed(self.seed)

        return random.random()


class ExampleNodeWithRandomNumberGen(Node):
    rnd: RandomNumberGen = zntrack.deps()

    outs: int = zntrack.outs()

    def run(self):
        self.outs = self.rnd.get_rnd()


def test_znodes_with_random_number_gen(proj_path):
    project = Project(force=True)

    rnd = RandomNumberGen()

    with project:
        node = ExampleNodeWithRandomNumberGen(rnd=rnd)
    project.run()
    # node.load(lazy=False)
    shutil.rmtree("nodes")

    project.run()

    node2 = node.from_rev()
    assert node.outs == node2.outs



import json
import pathlib

import yaml

import zntrack

CWD = pathlib.Path(__file__).parent.resolve()


class NodeFromCustomModule(zntrack.Node):
    _module_ = "zntrack.mymodule"

    def run(self):
        pass


def test__module_():
    proj_path = pathlib.Path.cwd()
    project = zntrack.Project()

    with project:
        NodeFromCustomModule()

    project.build()

    assert json.loads(
        (CWD / "zntrack_config" / "custom_module.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "custom_module.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "custom_module.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()



import pytest
import znflow

import zntrack


class CreateList(zntrack.Node):
    size: int = zntrack.params(10)

    output: list = zntrack.outs()

    def run(self):
        self.output = list(range(self.size))


class AddList(zntrack.Node):
    a: list = zntrack.deps()
    b: list = zntrack.deps()

    output: list = zntrack.outs()

    def run(self):
        self.output = self.a + self.b


@pytest.mark.parametrize("eager", [True, False])
def test_AddList(proj_path, eager):
    with zntrack.Project() as project:
        create_list_a = CreateList(name="CreateListA")
        create_list_b = CreateList(name="CreateListB")
        add_list = AddList(a=create_list_a.output, b=create_list_b.output)

    assert isinstance(add_list.a, znflow.Connection)
    assert isinstance(add_list.b, znflow.Connection)

    assert add_list.a.instance == create_list_a
    assert add_list.b.instance == create_list_b

    assert add_list.a.attribute == "output"
    assert add_list.b.attribute == "output"

    if eager:
        project.run()
    else:
        project.repro(build=True)

    assert create_list_a.output == list(range(10))
    assert create_list_b.output == list(range(10))
    assert add_list.output == list(range(10)) + list(range(10))


@pytest.mark.parametrize("eager", [True, False])
def test_AddList_getitem(proj_path, eager):
    with zntrack.Project() as project:
        create_list_a = CreateList(name="CreateListA")
        create_list_b = CreateList(name="CreateListB")
        add_list = AddList(a=create_list_a.output[1], b=create_list_b.output[-1])

    if eager:
        project.run()

        assert create_list_a.output == list(range(10))
        assert create_list_b.output == list(range(10))
        assert add_list.output == 10  # 1 + 9
    else:
        with pytest.raises(NotImplementedError):
            project.repro(build=True)


@pytest.mark.parametrize("eager", [True, False])
def test_AddList_slice(proj_path, eager):
    with zntrack.Project() as project:
        create_list_a = CreateList(name="CreateListA")
        create_list_b = CreateList(name="CreateListB")
        add_list = AddList(a=create_list_a.output[::2], b=create_list_b.output[2:7])

    if eager:
        project.run()

        assert create_list_a.output == list(range(10))
        assert create_list_b.output == list(range(10))
        assert add_list.output == list(range(10))[::2] + list(range(10))[2:7]
    else:
        with pytest.raises(NotImplementedError):
            project.repro(build=True)



import json
import pathlib

import pandas as pd
import yaml

import zntrack
import zntrack.group

CWD = pathlib.Path(__file__).parent.resolve()


class Plotting(zntrack.Node):
    a: pd.DataFrame = zntrack.plots(
        y="loss",
        x="epoch",
        x_label="Epoch",
        y_label="Loss",
        template="plotly_dark",
        title="Loss over Epochs",
    )


def test_plotting(proj_path):
    with zntrack.Project() as proj:
        Plotting()

    proj.build()

    assert json.loads((CWD / "zntrack_config" / "plots.json").read_text()) == json.loads(
        (proj_path / "zntrack.json").read_text()
    )
    assert yaml.safe_load(
        (CWD / "dvc_config" / "plots.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "plots.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_plotting("")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import zntrack

project = "ZnTrack"
copyright = "2025, Fabian Zills"
author = "Fabian Zills"
release = zntrack.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx.ext.viewcode",
    "sphinx_design",
    "sphinx.ext.autodoc",
    "sphinxcontrib.mermaid",
    "hoverxref.extension",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_favicon = "_static/favicon_ZnTrack.png"
# html_logo = "_static/logo_ZnTrack.png"
html_title = "ZnTrack"
html_static_path = ["_static"]
html_theme_options = {
    "navigation_with_keys": True,
    "source_repository": "https://github.com/zincware/ZnTrack/",
    "source_branch": "main",
    "source_directory": "docs/source/",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/zincware/ZnTrack",
            "html": "",
            "class": "fa-brands fa-github fa-2x",
        },
    ],
}

# font-awesome logos
html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/brands.min.css",
]

# -- Options for hoverxref extension -----------------------------------------
# https://sphinx-hoverxref.readthedocs.io/en/latest/

hoverxref_roles = ["term"]
hoverxref_role_types = {
    "class": "tooltip",
}

# -- Options for sphinx_copybutton -------------------------------------------
# https://sphinx-copybutton.readthedocs.io/en/latest/

copybutton_prompt_text = r">>> |\.\.\. |\(.*\) \$ "
copybutton_prompt_is_regexp = True



import json
import pathlib

import yaml

import zntrack.examples

IMPORT_URL_FILE = (
    "https://raw.githubusercontent.com/PythonFZ/zntrack-examples/refs/heads/main/LICENSE"
)

CWD = pathlib.Path(__file__).parent.resolve()


def test_add_url(proj_path) -> None:
    project = zntrack.Project()
    file = zntrack.add(IMPORT_URL_FILE, "LICENSE")

    with project:
        zntrack.examples.ReadFile(path=file)
        # zntrack.examples.ReadFile(path=file)
        # TODO: test with multiple nodes reading the file
        # TODO: test with repeated project.build() (after dvc repro)

    project.build()

    assert json.loads((CWD / "zntrack_config" / "add.json").read_text()) == json.loads(
        (proj_path / "zntrack.json").read_text()
    )
    assert yaml.safe_load(
        (CWD / "dvc_config" / "add.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "add.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()

    assert (proj_path / "LICENSE.dvc").exists()



"""Tests for 'zntrack.deps'-field which can be used as both
`zntrack.zn.deps` and `zntrack.zn.nodes`."""

import zntrack.examples

# TODO: change the parameters, rerun and see if it updates!


def test_as_deps(proj_path):
    """Test for 'zntrack.deps' acting as `zntrack.zn.deps`-like field."""
    project = zntrack.Project()

    with project:
        a = zntrack.examples.ComputeRandomNumber(params_file="a.json")
        b = zntrack.examples.ComputeRandomNumber(params_file="b.json")
        c = zntrack.examples.SumRandomNumbers(numbers=[a, b])

    project.build()

    a = a.from_rev(a.name)
    b = b.from_rev(b.name)

    a.write_params(min=1, max=5, seed=42)
    b.write_params(min=5, max=10, seed=42)

    project.repro(build=False)

    assert a.number == 1
    assert b.number == 10
    assert c.result == 11

    a.write_params(min=1, max=5, seed=31415)

    project.repro()

    a = a.from_rev(a.name)
    b = b.from_rev(b.name)
    c = c.from_rev(c.name)

    assert a.number == 5
    assert b.number == 10
    assert c.result == 15



import subprocess

import pytest
import znflow

import zntrack


class GenerateList(zntrack.Node):
    size: int = zntrack.params(10)
    outs: list = zntrack.outs()

    def run(self):
        self.outs = list(range(self.size))


class AddOneToList(zntrack.Node):
    data: list = zntrack.deps()
    outs: list = zntrack.outs()

    def run(self) -> None:
        assert isinstance(self.data, list)
        self.outs = [x + 1 for x in self.data]


class AddOneToDict(zntrack.Node):
    data: dict = zntrack.deps()
    outs: dict = zntrack.outs()

    def run(self) -> None:
        assert isinstance(self.data, dict)
        assert all(isinstance(v, list) for v in self.data.values())
        self.outs = {k: [x + 1 for x in v] for k, v in self.data.items()}


# TODO: combinedconnections need tests in files
@pytest.mark.parametrize("eager", [True, False])
def test_combined_connection(proj_path, eager):
    with zntrack.Project() as proj:
        a = GenerateList(size=1, name="a")
        b = GenerateList(size=2, name="b")
        c = GenerateList(size=3, name="c")

        data = a.outs + b.outs + c.outs
        added = AddOneToList(data=data)

    assert isinstance(data, znflow.CombinedConnections)
    proj.build()
    if eager:
        proj.run()
    else:
        subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

    assert added.outs == [1] + [1, 2] + [1, 2, 3]


@pytest.mark.parametrize("eager", [True, False])
def test_combine_dict(proj_path, eager):
    with zntrack.Project() as proj:
        a = GenerateList(size=1, name="a")
        b = GenerateList(size=2, name="b")
        c = GenerateList(size=3, name="c")

        added = AddOneToDict(data={x.name: x.outs for x in [a, b, c]})

    proj.build()
    if eager:
        proj.run()
    else:
        subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

    assert added.outs == {"a": [1], "b": [1, 2], "c": [1, 2, 3]}



import json
import pathlib

import yaml

import zntrack.examples

CWD = pathlib.Path(__file__).parent.resolve()


def test_metrics_as_deps(proj_path):
    project = zntrack.Project()

    with project:
        metrics_node = zntrack.examples.ParamsToMetrics(params={"loss": 0.01})

        zntrack.examples.DepsToMetrics(deps=metrics_node.metrics)
        zntrack.examples.DepsToMetrics(deps=metrics_node.params)

    project.build()

    assert json.loads(
        (CWD / "zntrack_config" / "metrics_deps.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "metrics_deps.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "metrics_deps.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()



import dataclasses

import zntrack


@dataclasses.dataclass
class SimpleThermostat:
    """Simple thermostat class"""


@dataclasses.dataclass
class ThermostatA:
    temperature: float = zntrack.params()


@dataclasses.dataclass
class ThermostatB:
    temperature: float = zntrack.params()


class MD(zntrack.Node):
    thermostat: ThermostatA | ThermostatB | SimpleThermostat = zntrack.deps()

    result: str = zntrack.outs()

    def run(self):
        self.result = self.thermostat.__class__.__name__


def test_switch_deps_class_keep_params(proj_path):
    """Test that changing the class triggers a DVC run."""
    proj = zntrack.Project()

    thermostat_a = ThermostatA(temperature=10)
    thermostat_b = ThermostatB(temperature=5)
    thermostat_c = SimpleThermostat()

    with proj:
        md = MD(thermostat=thermostat_a)

    proj.repro()
    assert md.from_rev().result == "ThermostatA"

    with proj:
        md.thermostat = thermostat_b

    proj.repro()
    assert md.from_rev().result == "ThermostatB"

    with proj:
        md.thermostat = thermostat_c

    proj.repro()
    assert md.from_rev().result == "SimpleThermostat"


if __name__ == "__main__":
    test_switch_deps_class_keep_params("")



import pytest

import zntrack.examples

apache_license = (
    "https://raw.githubusercontent.com/PythonFZ/zntrack-examples/refs/heads/main/LICENSE"
)

epl_license = "https://raw.githubusercontent.com/zincware/ZnDraw/refs/heads/main/LICENSE"


@pytest.mark.needs_internet
def test_add_read_url(proj_path) -> None:
    project = zntrack.Project()
    file = zntrack.add(apache_license, "LICENSE")

    with project:
        node_1 = zntrack.examples.ReadFile(path=file)
        node_2 = zntrack.examples.ReadFile(path=file)

    project.repro()

    assert "Apache License" in node_1.content
    assert "Apache License" in node_2.content

    project.repro()  # test run and build again


@pytest.mark.needs_internet
def test_add_two_files(proj_path) -> None:
    project = zntrack.Project()
    file_1 = zntrack.add(apache_license, "L1")
    file_2 = zntrack.add(apache_license, "L2")

    with project:
        node_1 = zntrack.examples.ReadFile(path=file_1)
        node_2 = zntrack.examples.ReadFile(path=file_2)

    project.repro()

    assert "Apache License" in node_1.content
    assert "Apache License" in node_2.content

    project.repro()  # test run and build again


@pytest.mark.needs_internet
def test_update_file(proj_path) -> None:
    project = zntrack.Project()
    file = zntrack.add(apache_license, "LICENSE")

    with project:
        node = zntrack.examples.ReadFile(path=file)

    project.repro()

    assert "Apache License" in node.content

    # new Project / rerunning

    project = zntrack.Project()
    file = zntrack.add(epl_license, "LICENSE", force=True)
    with project:
        node = zntrack.examples.ReadFile(path=file)

    project.repro()
    # TODO: can not use node.content because it seems to be read during building!

    assert "Eclipse Public License" in zntrack.from_rev(name=node.name).content

    project.repro()



import yaml

import zntrack
from zntrack import config


class NodeWithProperty(zntrack.Node):
    params: int = zntrack.params(None)

    @property
    def calc(self):
        """This should not change the params if not called."""
        self.params = 42
        return "calc"

    def run(self):
        pass


def test_NodeWithProperty(proj_path):
    with zntrack.Project() as proj:
        node = NodeWithProperty()

    proj.build()
    proj.run()

    node = node.from_rev()
    assert node.params is None


class NodeA(zntrack.Node):
    params1: int = zntrack.params(1)


class NodeB(zntrack.Node):
    params2: int = zntrack.params(2)


def test_SwitchNode(proj_path):
    with zntrack.Project() as proj:
        NodeA(name="Node")
    proj.build()

    params = yaml.safe_load(config.PARAMS_FILE_PATH.read_text())
    assert params["Node"] == {"params1": 1}

    with zntrack.Project() as proj:
        NodeB(name="Node")
    proj.build()

    params = yaml.safe_load(config.PARAMS_FILE_PATH.read_text())
    assert params["Node"] == {"params2": 2}


class CustomModule(zntrack.Node):
    _module_ = "package.module"


def test_CustomModule(proj_path):
    with zntrack.Project() as proj:
        CustomModule()
    proj.build()

    dvc = yaml.safe_load(config.DVC_FILE_PATH.read_text())
    assert (
        dvc["stages"]["CustomModule"]["cmd"]
        == "zntrack run package.module.CustomModule --name CustomModule"
    )



import pathlib

import pytest
import yaml

import zntrack
from zntrack.config import DVC_FILE_PATH


class SingleNode(zntrack.Node):
    params_file: pathlib.Path | list[pathlib.Path] = zntrack.params_path()

    def run(self):
        pass


def test_dvc_params(proj_path):
    """Test serialized data as parameters"""
    params = {"a": 100}
    file = pathlib.Path("my_params.yaml")

    file.write_text(yaml.safe_dump(params))
    with zntrack.Project() as proj:
        SingleNode(params_file=file)
    proj.build()
    proj.run()

    assert SingleNode.from_rev().params_file == file
    dvc_file = yaml.safe_load(DVC_FILE_PATH.read_text())
    assert [{"my_params.yaml": None}] == dvc_file["stages"]["SingleNode"]["params"]


def test_dvc_params_list(proj_path):
    """Test serialized data as parameters"""
    params = {"a": 100}
    file1 = pathlib.Path("my_params1.yaml")
    file2 = pathlib.Path("my_params2.yaml")

    file1.write_text(yaml.safe_dump(params))
    file2.write_text(yaml.safe_dump(params))

    with zntrack.Project() as proj:
        SingleNode(params_file=[file1, file2])
    proj.build()
    proj.run()

    assert SingleNode.from_rev().params_file == [file1, file2]
    dvc_file = yaml.safe_load(DVC_FILE_PATH.read_text())
    assert [{"my_params1.yaml": None}, {"my_params2.yaml": None}] == dvc_file["stages"][
        "SingleNode"
    ]["params"]


class MixedParams(zntrack.Node):
    params_file1: list[pathlib.Path] = zntrack.params_path()
    params_file2: pathlib.Path = zntrack.params_path()
    params: dict = zntrack.params()

    def run(self):
        pass


def test_dvc_mixed_params_list(proj_path):
    """Test serialized data as parameters"""
    params = {"a": 100}
    file1 = pathlib.Path("my_params1.yaml")
    file2 = pathlib.Path("my_params2.yaml")

    file1.write_text(yaml.safe_dump(params))
    file2.write_text(yaml.safe_dump(params))

    with zntrack.Project() as proj:
        MixedParams(params_file1=[file1, file2], params_file2=file1, params={"a": 100})
    proj.build()

    assert MixedParams.from_rev().params_file1 == [file1, file2]
    assert MixedParams.from_rev().params_file2 == file1
    assert MixedParams.from_rev().params == {"a": 100}
    dvc_file = yaml.safe_load(DVC_FILE_PATH.read_text())
    assert [
        "MixedParams",
        {"my_params1.yaml": None},
        {"my_params2.yaml": None},
    ] == dvc_file["stages"]["MixedParams"]["params"]


def test_dvc_params_error(proj_path):
    """Test serialized data as parameters"""
    with pytest.raises(ValueError):
        with zntrack.Project() as proj:
            SingleNode(params_file={"a": 100})
        proj.build()



import zntrack.examples


class NodeWithPostInit(zntrack.Node):
    def __post_init__(self):
        # The parent implements a __post_init__ method
        # this tests if super().__post_init__ is called correctly
        super().__post_init__()


def test_repr(proj_path):
    assert (
        repr(zntrack.examples.ParamsToOuts(params=42))
        == "ParamsToOuts(name='ParamsToOuts', params=42)"
    )
    assert repr(zntrack.examples.WriteDVCOuts(params=42)) == (
        "WriteDVCOuts(name='WriteDVCOuts', params=42,"
        " outs=PosixPath('nodes/WriteDVCOuts/output.txt'))"
    )
    assert repr(NodeWithPostInit()) == "NodeWithPostInit(name='NodeWithPostInit')"
    assert repr(zntrack.Node()) == "Node(name='Node')"
    assert repr(zntrack.Node(name="SomeNode")) == "Node(name='SomeNode')"
    assert (
        repr(zntrack.examples.ParamsToMetrics(params=42))
        == "ParamsToMetrics(name='ParamsToMetrics', params=42)"
    )
    assert (
        repr(zntrack.examples.WritePlots())
        == "WritePlots(name='WritePlots', x=[1, 2, 3], y=[4, 5, 6])"
    )


def test_repr_from_rev(proj_path):
    with zntrack.Project() as proj:
        n1 = zntrack.examples.ParamsToOuts(params=42)
        n2 = zntrack.examples.WriteDVCOuts(params=42)
        n6 = zntrack.examples.ParamsToMetrics(params=42)
        n7 = zntrack.examples.WritePlots()

    proj.build()

    assert repr(n1) == "ParamsToOuts(name='ParamsToOuts', params=42)"
    assert repr(n2) == (
        "WriteDVCOuts(name='WriteDVCOuts',"
        " params=42, outs=PosixPath('nodes/WriteDVCOuts/output.txt'))"
    )

    assert repr(n6) == "ParamsToMetrics(name='ParamsToMetrics', params=42)"
    assert repr(n7) == "WritePlots(name='WritePlots', x=[1, 2, 3], y=[4, 5, 6])"

    proj.run()

    assert repr(n1) == "ParamsToOuts(name='ParamsToOuts', params=42)"
    assert repr(n2) == (
        "WriteDVCOuts(name='WriteDVCOuts', params=42,"
        " outs=PosixPath('nodes/WriteDVCOuts/output.txt'))"
    )

    assert repr(n6) == "ParamsToMetrics(name='ParamsToMetrics', params=42)"
    assert repr(n7) == "WritePlots(name='WritePlots', x=[1, 2, 3], y=[4, 5, 6])"



import typing as t

import pytest

import zntrack
from zntrack import Node, Project


class InputsOutputs(Node):
    inputs: t.Any = zntrack.params()
    outputs: t.Any = zntrack.outs()


class WriteData(InputsOutputs):
    def run(self):
        self.outputs = self.inputs


class WriteDataWithInit(InputsOutputs):
    # def __init__(self, inputs=None, **kwargs):
    #     super().__init__(inputs=inputs, **kwargs)
    #     # this calls the auto_init of the subclass which demands the inputs argument!

    def run(self):
        self.outputs = self.inputs


class InOutsWInit(Node):
    inputs: t.Any = zntrack.params()
    outputs: t.Any = zntrack.outs()

    # def __init__(self, inputs=None, **kwargs):
    #     super().__init__(**kwargs)
    #     self.inputs = inputs


class WriteDataParentInit(InOutsWInit):
    def run(self):
        self.outputs = self.inputs


class WriteDataParentInitWithInit(InOutsWInit):
    # def __init__(self, inputs=None, **kwargs):
    #     super().__init__(**kwargs)
    #     self.inputs = inputs

    def run(self):
        self.outputs = self.inputs


@pytest.mark.parametrize(
    "cls",
    (WriteData, WriteDataWithInit, WriteDataParentInit, WriteDataParentInitWithInit),
)
def test_simple_inheritance(proj_path, cls):
    with Project() as project:
        node = cls(inputs="HelloWorld")
    project.run()
    # node.load()
    assert node.outputs == "HelloWorld"
    assert cls.from_rev().outputs == "HelloWorld"


class WriteCustomData(InputsOutputs):
    custom: str = zntrack.params()

    def run(self):
        self.outputs = f"{self.inputs} {self.custom}"


def test_WriteCustomData(proj_path):
    with Project() as project:
        node = WriteCustomData(inputs="Hello", custom="World")
    project.run()
    # node.load()
    assert node.outputs == "Hello World"
    assert WriteCustomData.from_rev().outputs == "Hello World"



"""Base class for zntrack plugins."""

import abc
import dataclasses
import pathlib
import typing as t

import yaml

from zntrack.config import (
    EXP_INFO_PATH,
    NOT_AVAILABLE,
    PLUGIN_EMPTY_RETRUN_VALUE,
    ZNTRACK_LAZY_VALUE,
)

if t.TYPE_CHECKING:
    from zntrack import Node


def plugin_getter(self: "Node", name: str):
    value = PLUGIN_EMPTY_RETRUN_VALUE

    field = self.state.get_field(name)

    for plugin in self.state.plugins.values():
        getter_value = plugin.getter(field)
        # TODO: is saving / loading part of a plugin or default feature?
        if getter_value is not PLUGIN_EMPTY_RETRUN_VALUE:
            if value is not PLUGIN_EMPTY_RETRUN_VALUE:
                raise ValueError(
                    "Multiple plugins return a value for "
                    f"{name}: {value} and {getter_value}"
                )
            value = getter_value
    return value


def _gitignore_file(path: str):
    """Add a path to the .gitignore file if it is not already there."""
    if not pathlib.Path(".gitignore").exists():
        with open(".gitignore", "w") as f:
            f.write(path + "\n")
        return
    with open(".gitignore", "r") as f:
        for line in f:
            if line.strip() == path:
                return
    with open(".gitignore", "a") as f:
        f.write(path + "\n")
    return


def get_exp_info() -> dict:
    if EXP_INFO_PATH.exists():
        return yaml.safe_load(EXP_INFO_PATH.read_text())
    return {}


def set_exp_info(data: dict) -> None:
    EXP_INFO_PATH.write_text(yaml.safe_dump(data))
    _gitignore_file(EXP_INFO_PATH.as_posix())


# TODO: have a dataclass for the base metrics, like hash, name, module, ...
@dataclasses.dataclass
class ZnTrackPlugin(abc.ABC):
    """ABC for writing zntrack plugins."""

    node: "Node"
    _continue_on_error_ = False

    @abc.abstractmethod
    def getter(self, field: dataclasses.Field) -> t.Any:
        """ZnField getter for zntrack options."""
        pass

    @abc.abstractmethod
    def save(self, field: dataclasses.Field) -> None:
        """Save method for zntrack options."""
        pass

    @abc.abstractmethod
    def convert_to_zntrack_json(self, graph) -> t.Any: ...

    @abc.abstractmethod
    def convert_to_dvc_yaml(self) -> t.Any: ...

    @abc.abstractmethod
    def convert_to_params_yaml(self) -> t.Any: ...

    def __enter__(self):
        self.setup()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def setup(self):
        return

    def close(self):
        return

    @classmethod
    def finalize(cls, **kwargs) -> None:
        return


def base_getter(
    self: "Node", name: str, func: t.Callable, suffix: t.Optional[str] = None
):
    if (
        name in self.__dict__
        and self.__dict__[name] is not ZNTRACK_LAZY_VALUE
        and self.__dict__[name] is not NOT_AVAILABLE
    ):
        return self.__dict__[name]

    if name in self.__dict__ and self.__dict__[name] is NOT_AVAILABLE:
        try:
            if suffix is not None:
                self.__dict__[name] = func(self, name, suffix)
            else:
                self.__dict__[name] = func(self, name)
        except FileNotFoundError:
            return NOT_AVAILABLE

    try:
        if suffix is not None:
            self.__dict__[name] = func(self, name, suffix)
        else:
            self.__dict__[name] = func(self, name)
    except FileNotFoundError:
        return NOT_AVAILABLE

    return getattr(self, name)



import os
import pathlib
import typing as t

import yaml
import znflow.utils

from zntrack.add import DVCImportPath
from zntrack.utils.import_handler import import_handler

from ..config import ENV_FILE_PATH


class RunDVCImportPathHandler(znflow.utils.IterableHandler):
    """Replace the nwd placeholder with the actual nwd."""

    def default(self, value):
        if isinstance(value, DVCImportPath):
            value.run()


def get_plugins_from_env(self):
    plugins_paths = os.environ.get(
        "ZNTRACK_PLUGINS", "zntrack.plugins.dvc_plugin.DVCPlugin"
    )
    plugins = [import_handler(p) for p in plugins_paths.split(",")]
    return {plugin.__name__: plugin(self) for plugin in plugins}


def get_attr_always_list(obj: t.Any, attr: str) -> list:
    value = getattr(obj, attr)
    if isinstance(value, dict):
        return list(value.values())
    if isinstance(value, tuple):
        return list(value)
    if not isinstance(value, list):
        value = [value]
    return value


def load_env_vars(name: str | None = None) -> None:
    # TODO: this should also use DVCFileSystem!
    if ENV_FILE_PATH.exists():
        env = yaml.safe_load(ENV_FILE_PATH.read_text())
        global_config = env.get("global", {})
        for key, val in global_config.items():
            if isinstance(val, list):
                global_config[key] = ",".join(val)
            elif isinstance(val, str):
                pass
            elif val is None:
                pass
            else:
                raise ValueError(
                    f"variable '{key}' has an invalid value '{val}' ({type(val)})"
                )
        os.environ.update(global_config)

        for key, value in env.get("stages", {}).get(name, {}).items():
            if isinstance(value, str):
                os.environ[key] = value
            elif isinstance(value, dict):
                os.environ.update(value)
            elif value is None:
                pass
            else:
                raise ValueError(
                    f"variable '{key}' has an invalid value '{val}' ({type(val)})"
                )


class TempPathLoader(znflow.utils.IterableHandler):
    def default(self, value, **kwargs):
        instance = kwargs["instance"]
        from zntrack.utils.node_wd import NWDReplaceHandler, get_nwd

        nwd_handler = NWDReplaceHandler()

        original_nwd = get_nwd(instance)
        tmp_nwd = instance.nwd

        original_path = pathlib.Path(nwd_handler(value, nwd=original_nwd))
        tmp_path = pathlib.Path(nwd_handler(value, nwd=tmp_nwd))

        if not tmp_path.exists():
            if instance.state.fs.isdir(original_path.as_posix()):
                instance.state.fs.get(
                    original_path.as_posix(),
                    instance.state.tmp_path.as_posix(),
                    recursive=True,
                )
            else:
                instance.state.fs.get(original_path.as_posix(), tmp_path.as_posix())

        if isinstance(value, pathlib.PurePath):
            return tmp_path
        else:
            return tmp_path.as_posix()


def sort_key(item):
    """Custom sorting key function to handle both string and dictionary types."""
    if isinstance(item, str):
        return item
    elif isinstance(item, dict):
        # For dictionaries, sort by their first (and only) key's string representation
        return list(item.keys())[0]


def sort_and_deduplicate(data: list[str | dict[str, dict]]):
    """Sort and deduplicate a list of strings and dictionaries."""
    data = sorted(data, key=sort_key)

    new_data = []
    for key in data:
        if key not in new_data:
            if isinstance(key, dict):
                if next(iter(key.keys())) in new_data:
                    raise ValueError(f"Found Duplicate key with different params: {key}")
                for other_key in new_data:
                    if isinstance(other_key, dict):
                        if next(iter(other_key.keys())) == next(iter(key.keys())):
                            if other_key != key:
                                raise ValueError(
                                    f"Found Duplicate key with different params: {key}"
                                )
            if isinstance(key, str):
                for other_key in new_data:
                    if isinstance(other_key, dict) and key in other_key.keys():
                        raise ValueError(
                            f"Found Duplicate key with different params: {key}"
                        )
            new_data.append(key)

    return new_data



"""Standard python init file for the utils directory."""

from zntrack.utils import cli

from .import_handler import module_handler
from .misc import get_attr_always_list
from .node_wd import NWDReplaceHandler, nwd

__all__ = ["module_handler", "nwd", "get_attr_always_list", "NWDReplaceHandler", "cli"]



import pathlib

import zntrack


class MyNode(zntrack.Node):
    def run(self):
        pass


def test_node_nwd(proj_path):
    with zntrack.Project():
        n = MyNode()

    assert n.nwd == pathlib.Path("nodes", "MyNode")


def test_no_node_name(proj_path):
    assert zntrack.Node().nwd == pathlib.Path("nodes", "Node")

    # it has to be given explicitly
    n = zntrack.Node(name="SomeNode")
    assert n.nwd == pathlib.Path("nodes", "SomeNode")



from zntrack import Node
from zntrack.examples.nodes import (
    AddNodeAttributes,
    AddNodeNumbers,
    AddNodes,
    AddNodes2,
    AddNumbers,
    AddNumbersProperty,
    AddOne,
    ComputeRandomNumber,
    ComputeRandomNumberNamed,
    ComputeRandomNumberWithParams,
    DepsToMetrics,
    NodeWithRestart,
    OptionalDeps,
    ParamsToMetrics,
    ParamsToOuts,
    ReadFile,
    SumNodeAttributes,
    SumNodeAttributesToMetrics,
    SumRandomNumbers,
    SumRandomNumbersNamed,
    WriteDVCOuts,
    WriteDVCOutsPath,
    WriteDVCOutsSequence,
    WriteMultipleDVCOuts,
    WritePlots,
)

__all__ = [
    "AddNumbers",
    "ReadFile",
    "ParamsToOuts",
    "ParamsToMetrics",
    "DepsToMetrics",
    "WritePlots",
    "AddNodeNumbers",
    "AddNumbersProperty",
    "AddNodes",
    "AddNodes2",
    "AddNodeAttributes",
    "AddNodeNumbers",
    "SumNodeAttributes",
    "SumNodeAttributesToMetrics",
    "AddOne",
    "WriteDVCOuts",
    "WriteDVCOutsSequence",
    "WriteDVCOutsPath",
    "WriteMultipleDVCOuts",
    "ComputeRandomNumber",
    "ComputeRandomNumberWithParams",
    "ComputeRandomNumberNamed",
    "SumRandomNumbers",
    "SumRandomNumbersNamed",
    "NodeWithRestart",
    "OptionalDeps",
]


def nodes() -> list[Node]:
    """Return the available nodes, grouped into categories."""

    return [globals()[name] for name in __all__]



import pandas as pd

from zntrack.config import NOT_AVAILABLE, ZNTRACK_OPTION_PLOTS_CONFIG, FieldTypes
from zntrack.fields.base import field
from zntrack.node import Node


def _plots_save_func(self: "Node", name: str, suffix: str):
    self.nwd.mkdir(parents=True, exist_ok=True)
    content = getattr(self, name)
    if not isinstance(content, pd.DataFrame):
        raise TypeError(f"Expected a pandas DataFrame, got {type(content)}")
    content.to_csv((self.nwd / name).with_suffix(suffix))


def _plots_autosave_setter(self: Node, name: str, value: pd.DataFrame):
    self.nwd.mkdir(parents=True, exist_ok=True)
    value.to_csv((self.nwd / name).with_suffix(".csv"))
    self.__dict__[name] = value


def _plots_getter(self: "Node", name: str, suffix: str):
    with self.state.fs.open((self.nwd / name).with_suffix(suffix)) as f:
        return pd.read_csv(f, index_col=0)


def plots(
    *,
    y: str | list[str] | None = None,
    cache: bool = True,
    independent: bool = False,
    x: str = "step",
    x_label: str | None = None,
    y_label: str | None = None,
    template: str | None = None,
    title: str | None = None,
    autosave: bool = False,
    **kwargs,
):
    """Pandas plot options.

    Parameters
    ----------
    y : str | list[str]
        Column name(s) to plot.
    cache : bool, optional
        Use the DVC cache, by default True.
    independent : bool, optional
        This fields output can be indepented of the
        input to the node. If set tue true, the
        entire Node output will be used for dependencies.
        Can be useful, if the output is e.g.
        a list of indices.
    x : str, optional
        Column name to use for the x-axis, by default "step".
    x_label : str, optional
        Label for the x-axis, by default None.
    y_label : str, optional
        Label for the y-axis, by default None.
    template : str, optional
        Plotly template to use, by default None.
    title : str, optional
        Title of the plot, by default None.
    autosave : bool, optional
        Save the data of this field every time it is being
        updated. Disable for large dataframes.

    Examples
    --------

    >>> import zntrack
    >>> import pandas as pd
    >>> class MyNode(zntrack.Node):
    ...     plots: pd.DataFrame = zntrack.plots(y="loss")
    ...
    ...     def run(self):
    ...         self.plots = pd.DataFrame({"loss": [1, 2, 3]})
    """
    if y is None:
        y = []

    kwargs["metadata"] = kwargs.get("metadata", {})

    plots_config = {}
    for key, value in {
        "x": x,
        "y": y,
        "x_label": x_label,
        "y_label": y_label,
        "template": template,
        "title": title,
    }.items():
        if value is not None:
            plots_config[key] = value
    if plots_config:
        kwargs["metadata"][ZNTRACK_OPTION_PLOTS_CONFIG] = plots_config

    if autosave:
        kwargs["setter"] = _plots_autosave_setter
    return field(
        default=NOT_AVAILABLE,
        cache=cache,
        independent=independent,
        field_type=FieldTypes.PLOTS,
        dump_fn=_plots_save_func,
        suffix=".csv",
        load_fn=_plots_getter,
        repr=False,
        init=False,
        **kwargs,
    )



import pytest
from dvc.stage.exceptions import InvalidStageName

import zntrack


class MyNode(zntrack.Node):
    pass


# Basic Naming Tests
def test_default_node_name(proj_path):
    """Test that a node gets the correct default name"""
    with zntrack.Project():
        node = MyNode()
        assert node.name == "MyNode"


def test_custom_node_name(proj_path):
    """Test that a node gets a custom name when specified"""
    with zntrack.Project():
        node = MyNode(name="CustomName")
        assert node.name == "CustomName"


# Duplicate Naming Tests
def test_unique_default_node_names(proj_path):
    """Test that multiple unnamed nodes receive unique names"""
    with zntrack.Project():
        n1 = MyNode()
        n2 = MyNode()
        n3 = MyNode()
        n4 = MyNode()

        assert n1.name == "MyNode"
        assert n2.name == "MyNode_1"
        assert n3.name == "MyNode_2"
        assert n4.name == "MyNode_3"


def test_duplicate_named_node_error(proj_path):
    """Test that an explicitly named node cannot be duplicated"""
    with zntrack.Project() as project:
        MyNode(name="A")

    with pytest.raises(ValueError, match="A node with the name 'A' already exists."):
        with project:
            MyNode(name="A")


# Grouped Naming Tests
def test_grouped_node_names(proj_path):
    """Test that nodes within groups get correct names"""
    project = zntrack.Project()

    with project:
        n1 = MyNode()
        n2 = MyNode()
        named_node = MyNode(name="SomeNode")

        assert n1.name == "MyNode"
        assert n2.name == "MyNode_1"
        assert named_node.name == "SomeNode"

    with project.group("grp1"):
        n3 = MyNode()
        n4 = MyNode()

        assert n3.name == "grp1_MyNode"
        assert n4.name == "grp1_MyNode_1"


def test_nested_grouped_node_names(proj_path):
    """Test that nested group names are correctly prefixed"""
    project = zntrack.Project()

    with project.group("A", "B"):
        n1 = MyNode()
        n2 = MyNode()

        assert n1.name == "A_B_MyNode"
        assert n2.name == "A_B_MyNode_1"


# Grouped Custom Name Tests
def test_grouped_custom_node_names(proj_path):
    """Test that nodes with custom names within groups get correctly prefixed"""
    project = zntrack.Project()

    with project.group("grp1"):
        n1 = MyNode(name="Alpha")
        n2 = MyNode(name="Beta")

        assert n1.name == "grp1_Alpha"
        assert n2.name == "grp1_Beta"

    project.build()

    assert n1.name == "grp1_Alpha"
    assert n2.name == "grp1_Beta"


def test_nested_grouped_custom_node_names(proj_path):
    """Test that nodes with custom names within nested groups get correctly prefixed"""
    project = zntrack.Project()

    with project.group("A", "B"):
        n1 = MyNode(name="Alpha")
        n2 = MyNode(name="Beta")

        assert n1.name == "A_B_Alpha"
        assert n2.name == "A_B_Beta"

    project.build()

    assert n1.name == "A_B_Alpha"
    assert n2.name == "A_B_Beta"
    assert n1.name == "A_B_Alpha"
    assert n2.name == "A_B_Beta"


# Grouped Duplicate Naming Tests
def test_grouped_duplicate_named_node(proj_path):
    """Test that nodes with the same custom name in different groups are unique"""
    project = zntrack.Project()

    with project.group("grp1"):
        n1 = MyNode(name="A")
        assert n1.name == "grp1_A"

    with project.group("grp2"):
        n2 = MyNode(name="A")
        assert n2.name == "grp2_A"

    project.build()

    assert n1.name == "grp1_A"
    assert n2.name == "grp2_A"
    assert n1.name == "grp1_A"
    assert n2.name == "grp2_A"

    with pytest.raises(ValueError, match="A node with the name 'grp1_A' already exists."):
        with project.group("grp1"):
            MyNode(name="A")


@pytest.mark.parametrize("char", ["@:", "#", "$", ":", "/", "\\", ".", ";", ","])
def test_forbidden_node_names(proj_path, char):
    """Test that nodes with forbidden names cannot be created"""
    with zntrack.Project():
        # https://github.com/iterative/dvc/blob/main/tests/func/test_run.py#L372-L375C32
        with pytest.raises(InvalidStageName):
            MyNode(name=f"copy_name-{char}")



import pandas as pd
import pytest

import zntrack


class Node(zntrack.Node):
    pass

    def run(self):
        pass


class NodeWithParameter(zntrack.Node):
    param: int = zntrack.params(default=1)

    def run(self):
        pass


class NodeWithOuts(zntrack.Node):
    out: int = zntrack.outs()

    def run(self):
        self.out = 1


class NodeWithMetrics(zntrack.Node):
    metric: dict = zntrack.metrics()

    def run(self):
        self.metric = {"y": 1}


class NodeWithPlots(zntrack.Node):
    plot: pd.DataFrame = zntrack.plots(y="y")

    def run(self):
        self.plot = pd.DataFrame({"y": [1]})


class NodeWithMetricsAndPlots(zntrack.Node):
    metric: dict = zntrack.metrics()
    plot: pd.DataFrame = zntrack.plots(y="y")

    def run(self):
        self.metric = {"y": 1}
        self.plot = pd.DataFrame({"y": [1]})


class NodeWithMetricsAndPlotsAndParams(zntrack.Node):
    metric: dict = zntrack.metrics()
    plot: pd.DataFrame = zntrack.plots(y="y")
    param: int = zntrack.params(default=1)

    def run(self):
        self.metric = {"y": 1}
        self.plot = pd.DataFrame({"y": [1]})


# TODO: test outside DVC repo


@pytest.mark.parametrize(
    ("node", "stage_hash", "full_stage_hash"),
    [
        (Node, "99a611464d", None),
        (NodeWithParameter, "8c8fb1673bb9", None),
        (NodeWithOuts, "77984ac3472e3", None),
        (NodeWithMetrics, "22877d83878b1", None),
        (NodeWithPlots, "e173cec6feabe", None),
        (NodeWithMetricsAndPlots, "a4b2aba9ec", None),
        (NodeWithMetricsAndPlotsAndParams, "b64bdd1da9", None),
    ],
)
def test_get_stage_hash(proj_path, node, stage_hash, full_stage_hash):
    with zntrack.Project() as proj:
        node = node()

    proj.build()

    assert node.state.get_stage_hash()[:10] == stage_hash[:10]

    proj.repro(build=False)

    assert node.from_rev().state.get_stage_hash()[:10] == stage_hash[:10]

    # TODO: this changes every run because of node_meta
    #  - do we want to exclude it - as it is the only output that is
    # by design non-deterministic!
    # assert node.from_rev().state.get_stage_hash(include_outs=True)[:10]
    #  == full_stage_hash[:10]



import json
import pathlib
import typing as t

import pandas as pd
import yaml

import zntrack

CWD = pathlib.Path(__file__).parent.resolve()


class NodeA1(zntrack.Node):
    metrics_paths: list[pathlib.Path] = zntrack.metrics_path()
    plots_path: list[pathlib.Path] = zntrack.plots_path()
    outs_path: pathlib.Path = zntrack.outs_path(zntrack.nwd / "my_output.txt")

    results: int = zntrack.outs()
    metrics: dict = zntrack.metrics()
    plots: pd.DataFrame = zntrack.plots(y="y")

    def run(self):
        pass


class NodeA2(zntrack.Node):
    outs_path: pathlib.Path = zntrack.outs_path(
        zntrack.nwd / "my_output.txt", independent=True
    )
    metrics_paths: list[pathlib.Path] = zntrack.metrics_path(independent=True)
    plots_path: list[pathlib.Path] = zntrack.plots_path(independent=True)

    results: int = zntrack.outs(independent=True)
    metrics: dict = zntrack.metrics(independent=True)
    plots: pd.DataFrame = zntrack.plots(independent=True, y="y")

    def run(self):
        pass


class NodeB(zntrack.Node):
    input: list[int] = zntrack.deps()

    def run(self):
        pass


class NodeC(zntrack.Node):
    deps: zntrack.Node | t.Any = zntrack.deps()

    def run(self):
        pass


class NodeWithProperty(zntrack.Node):
    @property
    def results(self):
        return 1

    def run(self):
        pass


def test_deps(proj_path):
    with zntrack.Project() as project:
        a = NodeA1(
            metrics_paths=[zntrack.nwd / "a.json", zntrack.nwd / "b.json"],
            plots_path=[zntrack.nwd / "p1.csv", zntrack.nwd / "p2.csv"],
        )
        b = NodeA2(
            metrics_paths=[zntrack.nwd / "a.json", zntrack.nwd / "b.json"],
            plots_path=[zntrack.nwd / "p1.csv", zntrack.nwd / "p2.csv"],
        )
        _ = NodeB(input=[a.results, b.results])

    with project.group("dependent"):
        _ = NodeC(deps=a)
        _ = NodeC(deps=a.results)
        _ = NodeC(deps=a.metrics)
        _ = NodeC(deps=a.plots)
        _ = NodeC(deps=a.outs_path)
        _ = NodeC(deps=a.metrics_paths)
        _ = NodeC(deps=a.plots_path)

    with project.group("independent"):
        _ = NodeC(deps=b)
        _ = NodeC(deps=b.results)
        _ = NodeC(deps=b.metrics)
        _ = NodeC(deps=b.plots)
        _ = NodeC(deps=b.outs_path)
        _ = NodeC(deps=b.metrics_paths)
        _ = NodeC(deps=b.plots_path)

    with project.group("property"):
        nwp = NodeWithProperty()
        _ = NodeC(deps=nwp.results)
        _ = NodeC(deps=nwp)

    project.build()

    assert json.loads(
        (CWD / "zntrack_config" / "dependencies.json").read_text()
    ) == json.loads((proj_path / "zntrack.json").read_text())
    assert yaml.safe_load(
        (CWD / "dvc_config" / "dependencies.yaml").read_text()
    ) == yaml.safe_load((proj_path / "dvc.yaml").read_text())
    assert (CWD / "params_config" / "dependencies.yaml").read_text() == (
        proj_path / "params.yaml"
    ).read_text()


if __name__ == "__main__":
    test_deps(None)



import numpy as np
import pytest

import zntrack


class ComputeA(zntrack.Node):
    """Node stage A"""

    inp: int = zntrack.params()
    out: float = zntrack.outs()

    def run(self):
        self.out = np.power(2, self.inp).item()


class ComputeB(zntrack.Node):
    """Node stage B"""

    inp: int = zntrack.params()
    out: float = zntrack.outs()

    def run(self):
        self.out = np.power(3, self.inp).item()


class ComputeAB(zntrack.Node):
    """Node stage AB, depending on A&B"""

    a: ComputeA = zntrack.deps()
    b: ComputeB = zntrack.deps()
    param: str = zntrack.params("default")
    out: float = zntrack.outs()

    def run(self):
        self.out = self.a.out + self.b.out


@pytest.mark.parametrize("eager", [True, False])
def test_stage_addition_run(proj_path, eager):
    """Adapted legacy tests"""
    with zntrack.Project() as project:
        a = ComputeA(inp=2)
        b = ComputeB(inp=3)
        ab = ComputeAB(a=a, b=b)

        a_named = ComputeA(inp=20)
        ab_named = ComputeAB(a=a_named, b=b)

    if eager:
        project.run()
    else:
        project.repro(build=True)

    assert ab.out == 31
    assert ab_named.out == 1048603



import pathlib
import subprocess

import zntrack.config
import zntrack.examples
import zntrack.exceptions


def test_load_WriteDVCOuts(proj_path):
    with zntrack.Project():
        node = zntrack.examples.WriteDVCOuts(params=42)

    assert node.__dict__["params"] == 42
    assert node.__dict__["outs"] == pathlib.Path("$nwd$/output.txt")

    assert node.params == 42
    assert node.outs == pathlib.Path("nodes/WriteDVCOuts/output.txt")

    assert node.__dict__["params"] == 42
    assert node.__dict__["outs"] == pathlib.Path("$nwd$/output.txt")


def test_load_ParamsToOuts(proj_path):
    with zntrack.Project() as project:
        node = zntrack.examples.ParamsToOuts(params=42)

    assert node.__dict__["params"] == 42
    assert "outs" not in node.__dict__

    assert node.params == 42
    assert node.outs == zntrack.config.NOT_AVAILABLE

    project.build()
    project.run()

    assert node.__dict__["params"] == 42
    assert node.__dict__["outs"] == 42

    assert node.params == 42
    assert node.outs == 42


def test_load_ParamsToOuts_from_rev(proj_path):
    with zntrack.Project() as project:
        zntrack.examples.ParamsToOuts(params=42)

    node = zntrack.examples.ParamsToOuts.from_rev()

    assert node.__dict__["params"] == zntrack.config.ZNTRACK_LAZY_VALUE
    assert "outs" not in node.__dict__

    assert node.params == zntrack.config.NOT_AVAILABLE
    assert node.outs == zntrack.config.NOT_AVAILABLE

    project.build()
    subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

    assert node.params == 42
    assert node.outs == 42


def test_load_WriteDVCOuts_from_rev(proj_path):
    with zntrack.Project() as project:
        zntrack.examples.WriteDVCOuts(params=42)

    node = zntrack.examples.WriteDVCOuts.from_rev()

    assert node.__dict__["params"] == zntrack.config.ZNTRACK_LAZY_VALUE
    # outs is actually outs_path!
    assert node.__dict__["outs"] == zntrack.config.ZNTRACK_LAZY_VALUE

    assert node.params == zntrack.config.NOT_AVAILABLE
    assert node.outs == zntrack.config.NOT_AVAILABLE

    project.build()
    subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

    assert node.params == 42
    assert node.outs == pathlib.Path("nodes/WriteDVCOuts/output.txt")



"""Zntrack apply module for custom "run" methods."""

import typing as t

o = t.TypeVar("o")


def apply(obj: o, method: str) -> o:
    """Update the default ``run`` method of ``zntrack.Node``.

    Parameters
    ----------
    obj : zntrack.Node
        The node to copy and update the ``run`` method.
    method : str
        The new method to use instead of the default ``run`` method.

    Returns
    -------
    zntrack.Node
        A new class which uses the new method instead of the default ``run`` method.

    Examples
    --------
    >>> import zntrack
    >>> class MyNode(zntrack.Node):
    ...     outs: str = zntrack.outs()
    ...
    ...     def run(self):
    ...         self.outs = "Hello, World!"
    ...
    ...     def my_run(self):
    ...         self.outs = "Hello, Custom World!"
    ...
    >>> OtherMyNode = zntrack.apply(MyNode, "my_run")
    >>> with zntrack.Project() as proj:
    ...     a = MyNode()
    ...     b = OtherMyNode()
    >>> proj.repro()
    >>> a.outs
    'Hello, World!'
    >>> b.outs
    'Hello, Custom World!'
    """

    if not hasattr(obj, method):
        raise AttributeError(f"The object does not have the requested method '{method}'.")

    class MockInheritanceClass(obj):
        """Copy of the original class with the new method attribute.

        We can not set the method directly on the original class, because
        it would be used by all the other instances of the class as well.
        """

        _method = method

    MockInheritanceClass.__module__ = obj.__module__
    MockInheritanceClass.__name__ = obj.__name__

    return MockInheritanceClass



import dataclasses

import dvc.api
import pytest
from fsspec.implementations.local import LocalFileSystem

import zntrack
from zntrack.config import NodeStatusEnum
from zntrack.state import NodeStatus


class MyNode(zntrack.Node):
    def run(self):
        pass


def test_state_get(proj_path):
    with zntrack.Project() as project:
        n = MyNode()

    project.build()
    assert isinstance(n.state, NodeStatus)
    assert n.state.remote is None
    assert n.state.rev is None
    assert n.state.run_count == 0
    assert not n.state.restarted
    assert n.state.state == NodeStatusEnum.CREATED
    assert n.name == "MyNode"
    assert n.state.name == "MyNode"

    assert isinstance(n.state.fs, LocalFileSystem)

    n = n.from_rev(remote=".")  # fake a remote by using the current directory
    assert n.state.remote == "."
    assert isinstance(n.state.fs, dvc.api.DVCFileSystem)


def test_state_get_after_run(proj_path):
    with zntrack.Project() as project:
        n = MyNode()

    project.build()
    assert n.state.get_stage().addressing == "MyNode"
    assert n.state.state == NodeStatusEnum.CREATED
    project.repro(build=False)

    assert n.state.get_stage().addressing == "MyNode"
    assert (
        n.state.get_stage_lock()["cmd"]
        == "zntrack run test_node_state.MyNode --name MyNode"
    )
    n = n.from_rev()
    assert "deps" not in n.state.get_stage_lock()
    assert len(n.state.get_stage_lock()["outs"]) == 1
    assert n.state.get_stage_lock()["outs"][0]["hash"] == "md5"
    assert n.state.get_stage_lock()["outs"][0]["path"] == "nodes/MyNode/node-meta.json"
    assert n.state.get_stage_lock()["outs"][0]["size"] > 100  # run time can vary
    assert "md5" in n.state.get_stage_lock()["outs"][0]
    assert n.state.remote is None
    assert n.state.rev is None
    assert n.state.run_count == 1
    assert not n.state.restarted
    assert n.state.state == NodeStatusEnum.FINISHED
    assert n.state.lazy_evaluation is True

    # now test the loaded node
    n = n.from_rev()

    assert n.state.remote is None
    assert n.state.rev is None
    assert n.state.run_count == 1
    assert not n.state.restarted
    assert n.state.state == NodeStatusEnum.FINISHED
    assert n.state.lazy_evaluation is True

    # again just for lazy_evaluation=False
    n = n.from_rev(lazy_evaluation=False)
    assert n.state.lazy_evaluation is False


def test_state_set(proj_path):
    with zntrack.Project() as project:
        n = MyNode()

    project.build()
    with pytest.raises(dataclasses.FrozenInstanceError):
        n.state.remote = "remote"
    with pytest.raises(dataclasses.FrozenInstanceError):
        n.state.rev = "rev"
    with pytest.raises(dataclasses.FrozenInstanceError):
        n.state.run_count = 1
    with pytest.raises(dataclasses.FrozenInstanceError):
        n.state.state = NodeStatusEnum.RUNNING
    with pytest.raises(dataclasses.FrozenInstanceError):
        n.state.restarted = True
    with pytest.raises(dataclasses.FrozenInstanceError):
        n.state.lazy_evaluation = False



import subprocess

import pandas as pd

import zntrack


class PandasPlotNode(zntrack.Node):
    n: int = zntrack.params()

    plot: pd.DataFrame = zntrack.plots(y="y", x="x")

    def run(self):
        self.plot = pd.DataFrame({"x": range(self.n), "y": range(self.n)})


class AutoSavePandasPlotNode(zntrack.Node):
    n: int = zntrack.params()

    plot: pd.DataFrame = zntrack.plots(y="y", x="x", autosave=True)

    def run(self):
        self.plot = pd.DataFrame({"x": [], "y": []})
        for i in range(self.n):
            self.plot = pd.concat([self.plot, pd.DataFrame({"x": [i], "y": [i]})])
            with (self.nwd / "plot.csv").open("r") as f:
                df = pd.read_csv(f, index_col=0)
                assert len(df) == i + 1
                assert len(self.plot) == i + 1
                assert df.equals(self.plot)


def test_simple_plot(proj_path):
    with zntrack.Project() as proj:
        node = PandasPlotNode(n=10)

    proj.build()
    subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

    df = pd.read_csv(node.nwd / "plot.csv")
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 10

    assert isinstance(node.plot, pd.DataFrame)
    assert len(node.plot) == 10

    # create a new instance (lazy loading)
    node = node.from_rev()
    assert isinstance(node.plot, pd.DataFrame)
    assert len(node.plot) == 10


def test_multiple_plots(proj_path):
    with zntrack.Project() as proj:
        n1 = PandasPlotNode(n=10)
        n2 = PandasPlotNode(n=20)
        n3 = PandasPlotNode(n=30)

    proj.build()
    subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

    for idx, node in enumerate([n1, n2, n3]):
        df = pd.read_csv(node.nwd / "plot.csv")
        assert isinstance(df, pd.DataFrame)
        assert len(df) == 10 * (idx + 1)

        assert isinstance(node.plot, pd.DataFrame)
        assert len(node.plot) == 10 * (idx + 1)

        # create a new instance (lazy loading)
        node = node.from_rev(node.name)
        assert isinstance(node.plot, pd.DataFrame)
        assert len(node.plot) == 10 * (idx + 1)


def test_groups(proj_path):
    project = zntrack.Project()
    with project.group("group1"):
        n1 = PandasPlotNode(n=10)
        n2 = PandasPlotNode(n=20)
    with project.group("group2"):
        n3 = PandasPlotNode(n=30)
        n4 = PandasPlotNode(n=40)

    project.build()
    subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)

    for idx, node in enumerate([n1, n2, n3, n4]):
        df = pd.read_csv(node.nwd / "plot.csv")
        assert isinstance(df, pd.DataFrame)
        assert len(df) == 10 * (idx + 1)

        assert isinstance(node.plot, pd.DataFrame)
        assert len(node.plot) == 10 * (idx + 1)

        # create a new instance (lazy loading)
        node = node.from_rev(node.name)
        assert isinstance(node.plot, pd.DataFrame)
        assert len(node.plot) == 10 * (idx + 1)


def test_autosave(proj_path):
    with zntrack.Project() as proj:
        AutoSavePandasPlotNode(n=10)

    proj.build()
    subprocess.run(["dvc", "repro"], cwd=proj_path, check=True)



import zntrack


class NodeWithPostLoad(zntrack.Node):
    params: float = zntrack.params()
    outs: float = zntrack.outs()

    def _post_load_(self):
        self.params += 1

    def run(self):
        self.outs = self.params


class DepsOnPostLoadNode(zntrack.Node):
    deps: float = zntrack.deps()
    outs: float = zntrack.outs()

    def run(self):
        self.outs = self.deps


def test_post_load(proj_path):
    project = zntrack.Project()
    with project:
        node = NodeWithPostLoad(params=1)

    assert node.params == 1
    project.repro()

    n = node.from_rev()
    assert n.params == 2  # modified by _post_load_
    assert n.outs == 1


def test_post_load_deps(proj_path):
    project = zntrack.Project()

    with project:
        node = NodeWithPostLoad(params=1)
        dep_node = DepsOnPostLoadNode(deps=node.outs)
        dep_node_2 = DepsOnPostLoadNode(deps=node.params)

    project.repro()

    # post load has not been called
    assert dep_node.outs == 1
    assert node.outs == 1
    assert node.params == 1
    # assert dep_node_2.deps == 2 # has not been resolved from a connection
    assert dep_node_2.outs == 2

    # post load has been called
    n = dep_node.from_rev()
    assert n.outs == 1
    assert n.deps == 1

    # _post_load_ will also be called if the node is resolved from a connection
    n2 = dep_node_2.from_rev(name=dep_node_2.name)
    assert n2.outs == 2
    assert n2.deps == 2

    assert node.from_rev().outs == 1
    assert node.from_rev().params == 2



import pytest

import zntrack.examples


@pytest.mark.xfail(reason="pending implementation")
@pytest.mark.needs_internet
def test_get_nodes_remote(proj_path):
    nodes = zntrack.get_nodes(
        remote="https://github.com/PythonFZ/ZnTrackExamples.git", rev="890c714"
    )

    assert nodes["HelloWorld"].random_number == 123


@pytest.mark.xfail(reason="pending implementation")
def test_get_nodes(proj_path):
    with zntrack.Project(automatic_node_names=True) as proj:
        _ = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    with proj.group("example1"):
        _ = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    with proj.group("nested", "GRP1"):
        _ = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)
    with proj.group("nested", "GRP2"):
        _ = zntrack.examples.ParamsToOuts(params=15)
        _ = zntrack.examples.ParamsToOuts(params=15)

    proj.run()

    nodes = zntrack.get_nodes(remote=proj_path, rev=None)

    assert nodes["ParamsToOuts"].outs == 15
    assert nodes["ParamsToOuts_1"].outs == 15
    assert nodes["example1_ParamsToOuts"].outs == 15
    assert nodes["example1_ParamsToOuts_1"].outs == 15
    assert nodes["nested_GRP1_ParamsToOuts"].outs == 15
    assert nodes["nested_GRP1_ParamsToOuts_1"].outs == 15
    assert nodes["nested_GRP2_ParamsToOuts"].outs == 15
    assert nodes["nested_GRP2_ParamsToOuts_1"].outs == 15



import dataclasses
import importlib
import pathlib
import subprocess
import typing as t
import warnings

import yaml
import znflow
import znjson

from zntrack.add import DVCImportPath
from zntrack.config import (
    FIELD_TYPE,
    PARAMS_FILE_PATH,
    ZNTRACK_FIELD_SUFFIX,
    ZNTRACK_INDEPENDENT_OUTPUT_TYPE,
    FieldTypes,
)

from .node import Node
from .utils import module_handler


class DataclassContainer:
    def __init__(self, cls):
        self.cls = cls

    def get_with_params(self, node_name, attr_name, index: int | None = None):
        """Get an instance of the dataclass with the parameters from the params file.

        Attributes
        ----------
        node_name : str
            The name of the node.
        attr_name : str
            The name of the attribute.
        index : int | None
            The index in the parameter file, if the node attribute
            is a list of dataclasses. None if a single dataclass.

        """
        all_params = yaml.safe_load(PARAMS_FILE_PATH.read_text())
        if index is not None:
            dc_params = all_params[node_name][attr_name][index]
        else:
            dc_params = all_params[node_name][attr_name]
        dc_params.pop("_cls", None)
        return self.cls(**dc_params)


def _enforce_str_list(content) -> list[str]:
    if isinstance(content, (str, pathlib.Path)):
        return [pathlib.Path(content).as_posix()]
    elif isinstance(content, (list, tuple)):
        return [pathlib.Path(x).as_posix() for x in content]
    else:
        raise ValueError(f"found unsupported content type '{content}'")


class NodeDict(t.TypedDict):
    module: str
    name: str
    cls: str
    remote: t.Optional[t.Any]
    rev: t.Optional[t.Any]


class NodeConverter(znjson.ConverterBase):
    level = 100
    instance = Node
    representation = "zntrack.Node"

    def encode(self, obj: Node) -> NodeDict:
        return {
            "module": module_handler(obj),
            "name": obj.name,
            "cls": obj.__class__.__name__,
            "remote": obj.state.remote,
            "rev": obj.state.rev,
        }

    def decode(self, s: dict) -> Node:
        module = importlib.import_module(s["module"])
        cls = getattr(module, s["cls"])
        return cls.from_rev(name=s["name"], remote=s["remote"], rev=s["rev"])


class ConnectionConverter(znjson.ConverterBase):
    """Convert a znflow.Connection object to dict and back."""

    level = 100
    representation = "znflow.Connection"
    instance = znflow.Connection

    def encode(self, obj: znflow.Connection) -> dict:
        """Convert the znflow.Connection object to dict."""
        if obj.item is not None:
            raise NotImplementedError("znflow.Connection getitem is not supported yet.")
        # Can not use `dataclasses.asdict` because it automatically
        # converts nested dataclasses to dict.
        return {
            "instance": obj.instance,
            "attribute": obj.attribute,
            "item": obj.item,
        }

    def decode(self, value: dict) -> znflow.Connection:
        """Create znflow.Connection object from dict."""
        return znflow.Connection(**value)


class CombinedConnectionsConverter(znjson.ConverterBase):
    """Convert a znflow.Connection object to dict and back."""

    level = 100
    representation = "znflow.CombinedConnections"
    instance = znflow.CombinedConnections

    def encode(self, obj: znflow.CombinedConnections) -> dict:
        """Convert the znflow.Connection object to dict."""
        if obj.item is not None:
            raise NotImplementedError(
                "znflow.CombinedConnections getitem is not supported yet."
            )
        return {
            "connections": obj.connections,
            "item": obj.item,
        }

    def decode(self, value: dict) -> znflow.CombinedConnections:
        """Create znflow.Connection object from dict."""
        return znflow.CombinedConnections(**value)


def node_to_output_paths(node: Node, attribute: str) -> t.List[str]:
    """Get all output paths for a node."""
    # TODO: this should be a part of the DVCPlugin!
    if not isinstance(node, Node):
        raise ValueError(f"Expected a Node object, got {type(node)}")
    import_path = None
    if node._external_:
        # DODO: use the dvc_stage hash instead!
        # try use the one including the outputs
        import_path = pathlib.Path(
            "external", node.state.get_stage_hash(include_outs=True)[:32]
        )
        import_path.mkdir(exist_ok=True, parents=True)
        # we need to make the parent directory of the output
        # that directory is probably best described by using the node.name
        # of the node that depends on the import?
        # or we use a hash from commit / node name / repo path <-- only validate answer!
        # we want to run dvc import remote get_path(node, "attribute")
        # --rev rev --out /.../get_path(node, "attribute").name
        # use --no-download option while building
        # check how dvc repro or paraffin would download files?
        # Do we want the user to force download?
        # have zntrack.Path(path, remote, rev, is_dvc_tracked, is_db)
        #  to use dvc import-url / import-db in the graph
        # return []
        # raise NotImplementedError
    if attribute is None:
        fields = dataclasses.fields(node)
    else:
        try:
            fields = [node.state.get_field(attribute)]
        except AttributeError:
            # if you e.g. pass a property, we can not
            # determine what data is used and need
            # to assume all fields are used.
            # TODO: tests
            # TODO: have a custom decorator which defines the fields used?
            fields = dataclasses.fields(node)
    paths = []
    for field in fields:
        option_type = field.metadata.get(FIELD_TYPE)

        if any(
            option_type is x
            for x in [
                FieldTypes.PARAMS,
                FieldTypes.PARAMS,
                FieldTypes.DEPS,
                FieldTypes.DEPS_PATH,
                None,
            ]
        ):
            continue
        if node._external_:
            warnings.warn("External nodes are currently always loaded dynamically.")
            continue
        if field.metadata.get(ZNTRACK_INDEPENDENT_OUTPUT_TYPE) is True:
            paths.append((node.nwd / "node-meta.json").as_posix())
            if node._external_:
                raise NotImplementedError
                # paths.append((import_path / "node-meta.json").as_posix())
        if option_type in [
            FieldTypes.OUTS,
            FieldTypes.PLOTS,
            FieldTypes.METRICS,
        ]:
            suffix = field.metadata[ZNTRACK_FIELD_SUFFIX]
            paths.append((node.nwd / field.name).with_suffix(suffix).as_posix())
        elif option_type == FieldTypes.OUTS_PATH:
            paths.extend(_enforce_str_list(getattr(node, field.name)))
        elif option_type == FieldTypes.PLOTS_PATH:
            paths.extend(_enforce_str_list(getattr(node, field.name)))
        elif option_type == FieldTypes.METRICS_PATH:
            paths.extend(_enforce_str_list(getattr(node, field.name)))

    if len(paths) == 0:
        if node._external_:
            node_meta_path = import_path / "node-meta.json"
            if not node_meta_path.exists():
                cmd = [
                    "dvc",
                    "import",
                    node.state.remote,
                    (node.nwd / "node-meta.json").as_posix(),
                ]
                if node.state.rev is not None:
                    cmd.extend(["--rev", node.state.rev])
                cmd.append("--no-download")  # consider using --no-exec
                cmd.extend(["--out", node_meta_path.as_posix()])
                subprocess.check_call(cmd)
            paths.append(node_meta_path.as_posix())
        else:
            # for nodes with no outputs, we rely on 'node-meta.json'
            paths.append((node.nwd / "node-meta.json").as_posix())

    return paths


class DataclassConverter(znjson.ConverterBase):
    """Convert a python dataclass object to dict and back.

    This converter does not return an instance, but
    only the class inside a DataclassContainer.
    Saving the values must be done separately.

    Attributes
    ----------
    representation: str
        representation inside the dict for deserialization.
    level: int
        The level in which the encoding should be applied. A higher number means it will
        try this first. E.g. test small numpy conversion before pickle
        first.

    """

    level = 20
    representation = "@dataclasses.dataclass"

    def encode(self, obj: object) -> dict:
        """Convert the znflow.Connection object to dict."""
        module = module_handler(obj)
        cls = obj.__class__.__name__

        return {
            "module": module,
            "cls": cls,
        }

    def decode(self, value: dict) -> DataclassContainer:
        """Create znflow.Connection object from dict."""
        module = importlib.import_module(value["module"])
        cls = getattr(module, value["cls"])
        return DataclassContainer(cls)

    def __eq__(self, other) -> bool:
        if dataclasses.is_dataclass(other) and not isinstance(
            other, (Node, znflow.Connection, znflow.CombinedConnections)
        ):
            return True
        return False


class DVCImportPathConverter(znjson.ConverterBase):
    """Convert a DVCImportPath object to to pathlib.

    We do not want to store the information about the DVCImportPath
    as this is handled in the "path.dvc" file.
    Thus, we just make use of the pathlib.Path representation.
    """

    level = 100
    instance = DVCImportPath
    representation = "pathlib.Path"

    def encode(self, obj: DVCImportPath) -> str:
        return obj.path.as_posix()

    def decode(self, value: str) -> None:
        raise NotImplementedError("DVCImportPath is converted to pathlib.Path.")



import dataclasses
import functools
import typing as t

import znfields

from zntrack.config import (
    FIELD_TYPE,
    ZNTRACK_CACHE,
    ZNTRACK_FIELD_DUMP,
    ZNTRACK_FIELD_LOAD,
    ZNTRACK_FIELD_SUFFIX,
    ZNTRACK_INDEPENDENT_OUTPUT_TYPE,
    FieldTypes,
)
from zntrack.node import Node
from zntrack.plugins import base_getter, plugin_getter

FN_WITH_SUFFIX = t.Callable[["Node", str, str], t.Any]
FN_WITHOUT_SUFFIX = t.Callable[["Node", str], t.Any]


def field(
    default=dataclasses.MISSING,
    *,
    cache: bool = True,
    independent: bool = False,
    field_type: FieldTypes,
    dump_fn: FN_WITH_SUFFIX | FN_WITHOUT_SUFFIX | None = None,
    suffix: str | None = None,
    load_fn: FN_WITHOUT_SUFFIX | FN_WITH_SUFFIX | None = None,
    **kwargs,
):
    """Create a custom field.

    Arguments
    ---------
    default : t.Any
        Default value of the field.
        For an output field, this should be ``zntrack.NOT_AVAILABLE``
        and should not be set during initialization.
    cache : bool
        Use the DVC cache for the field.
    independent : bool
        If the output of this field can be independent of the input.
    field_type : FieldTypes
        The type of the field.
    dump_fn : FN_WITH_SUFFIX | FN_WITHOUT_SUFFIX
        Function to dump the field.
    suffix : str
        Suffix to append to the field name.
        Can be None if the output is a directory.
    load_fn : FN_WITHOUT_SUFFIX | FN_WITH_SUFFIX
        Function to load the field.
    **kwargs
        Additional arguments to pass to the field.

    Examples
    --------

    >>> import numpy as np
    >>> import zntrack
    ...
    >>>    def _load_fn(self: zntrack.Node, name: str, suffix: str) -> np.ndarray:
    ...        with self.state.fs.open(
    ...            (self.nwd / name).with_suffix(suffix), mode="rb"
    ...        ) as f:
    ...            return np.load(f)
    ...
    >>> def _dump_fn(self: zntrack.Node, name: str, suffix: str) -> None:
    ...     with open((self.nwd / name).with_suffix(suffix), mode="wb") as f:
    ...         np.save(f, getattr(self, name))
    ...
    >>> def numpy_field(*, cache: bool = True, independent: bool = False, **kwargs):
    ...     return field(
                default=zntrack.NOT_AVAILABLE
    ...         cache=cache,
    ...         independent=independent,
    ...         field_type=zntrack.FieldTypes.OUTS,
    ...         dump_fn=_dump_fn,
    ...         suffix=".npy",
    ...         load_fn=_load_fn,
    ...         **kwargs,
    ...     )
    ...
    >>> class MyNode(Node):
    ...     data: np.ndarray = numpy_field()
    ...
    ...     def run(self) -> None:
    ...         self.data = np.arange(9).reshape(3, 3)
    """
    kwargs["metadata"] = kwargs.get("metadata", {})
    kwargs["metadata"][FIELD_TYPE] = field_type
    kwargs["metadata"][ZNTRACK_CACHE] = cache
    kwargs["metadata"][ZNTRACK_INDEPENDENT_OUTPUT_TYPE] = independent
    if load_fn is not None:
        kwargs["metadata"][ZNTRACK_FIELD_LOAD] = functools.partial(
            base_getter, func=load_fn
        )
    if dump_fn is not None:
        kwargs["metadata"][ZNTRACK_FIELD_DUMP] = dump_fn
    if suffix is not None:
        kwargs["metadata"][ZNTRACK_FIELD_SUFFIX] = suffix
    return znfields.field(default=default, getter=plugin_getter, **kwargs)



import functools

import zntrack.examples


class NodeWithProperty(zntrack.Node):
    params: int = zntrack.params(42)
    outs: dict = zntrack.outs()

    def run(self):
        self.outs = {"outs": self.params}

    @functools.cached_property
    def value(self):
        self.property_triggered = True
        return self.outs


def test_property_access_graph_building(proj_path):
    project = zntrack.Project()

    with project:
        node = NodeWithProperty()
        zntrack.examples.DepsToMetrics(deps=node.value)

    project.build()

    assert not hasattr(node, "property_triggered")

    project.repro()

    assert not hasattr(node, "property_triggered")

    # assert node.outs != -5

    with project:
        node = NodeWithProperty()
        zntrack.examples.DepsToMetrics(deps=node.value)

    assert not hasattr(node, "property_triggered")
    project.build()
    assert not hasattr(node, "property_triggered")

    # sanity check
    node.value  # trigger property access
    assert node.property_triggered is True



import importlib
import logging
import pathlib
import sys
import typing as t

if t.TYPE_CHECKING:
    from ..node import Node

log = logging.getLogger(__name__)


def import_handler(node_path: str) -> t.Type["Node"]:
    """Import a module from a string.

    node_path : str
        The full path to the Node, e.g. `ipsuite.nodes.SmilesToAtoms`
    """
    module_path, class_name = node_path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    cls = getattr(module, class_name)
    return cls


def module_handler(obj) -> str:
    """Get the module for the Node.

    There are three cases that have to be handled here:
        1. Run from __main__ should not have __main__ as module but
            the actual filename.
        2. Run from a Jupyter Notebook should not return the launchers name
            but __main__ because that might be used in tests
        3. Return the plain module if the above are not fulfilled.

    Parameters
    ----------
    obj:
        Any object that implements __module__

    """
    # if config.nb_name:
    #     try:
    #         return f"{config.nb_class_path}.{obj.__name__}"
    #     except AttributeError:
    #         return f"{config.nb_class_path}.{obj.__class__.__name__}"
    if obj.__module__ != "__main__":
        if hasattr(obj, "_module_"):  # allow module override
            return obj._module_
        return obj.__module__
    if pathlib.Path(sys.argv[0]).stem == "ipykernel_launcher":
        # special case for e.g. testing
        return obj.__module__
    return pathlib.Path(sys.argv[0]).stem



import subprocess

import pytest

import zntrack.examples


class NodeInGroup(zntrack.Node):
    def run(self):
        assert self.state.group.names == ("MyGrp",)


def test_groups_io(proj_path):
    project = zntrack.Project()

    with project:
        a = zntrack.examples.ParamsToOuts(params=1)

    with project.group("A"):
        b = zntrack.examples.AddOne(number=a.outs)

    with project.group("A"):
        c = zntrack.examples.AddOne(number=b.outs)

    with project.group("A", "B"):
        d = zntrack.examples.AddOne(number=c.outs)

    project.build()
    subprocess.check_call(["dvc", "repro"])

    assert a.state.group is None
    assert b.state.group.names == ("A",)
    assert c.state.group == b.state.group
    assert d.state.group.names == ("A", "B")

    assert a.outs == 1
    assert b.outs == 2
    assert c.outs == 3
    assert d.outs == 4

    assert a.name == "ParamsToOuts"
    assert b.name == "A_AddOne"
    assert c.name == "A_AddOne_1"
    assert d.name == "A_B_AddOne"

    assert len(b.state.group) == 2
    assert a not in b.state.group
    assert b in b.state.group
    assert c in b.state.group
    assert d not in b.state.group

    assert b.state.group["A_AddOne"] == b

    with pytest.raises(KeyError):
        b.state.group["NotExisting"]

    for node in d.state.group:
        assert node == d

    with pytest.raises(AttributeError):
        b.state.group.names = "Hello"

    with pytest.raises(AttributeError):
        b.state.group.nodes = []


def test_nested_groups(proj_path):
    # disabled from within znflow
    project = zntrack.Project()

    with project.group("A"):
        with pytest.raises(TypeError):
            with project.group("B"):
                pass


def test_node_in_group(proj_path):
    project = zntrack.Project()
    with project.group("MyGrp"):
        n = NodeInGroup()

    project.repro()
    assert n.state.group.names == ("MyGrp",)


def test_custom_node_name_in_group(proj_path):
    project = zntrack.Project()
    with project.group("MyGrp"):
        n = NodeInGroup(name="CustomName")
        m = NodeInGroup()
        assert m.name == "MyGrp_NodeInGroup"
        assert n.name == "MyGrp_CustomName"

    project.repro()

    assert m.name == "MyGrp_NodeInGroup"
    assert m.state.group.names == ("MyGrp",)

    assert n.name == "MyGrp_CustomName"
    assert n.state.group.names == ("MyGrp",)


if __name__ == "__main__":
    test_groups_io(None)



import time
import typing as t

import pytest

import zntrack

# TODO Add ZNTRACK.TOOLS.TIMEIT


class SleepClassNoMetadata(zntrack.Node):
    sleep_for: int = zntrack.params(1)

    @zntrack.tools.timeit("metadata")
    def run(self):
        time.sleep(self.sleep_for)


class SleepClass(zntrack.Node):
    sleep_for: int = zntrack.params(1)
    timeit_metrics: t.Any = zntrack.metrics()

    @zntrack.tools.timeit("timeit_metrics")
    def run(self):
        time.sleep(self.sleep_for)


class SleepClassLoop(zntrack.Node):
    sleep_for: float = zntrack.params(0.1)
    timeit_metrics: t.Any = zntrack.metrics()
    iterations: int = zntrack.params(30)

    def run(self):
        for _ in range(self.iterations):
            self.sleep()

    @zntrack.tools.timeit("timeit_metrics")
    def sleep(self):
        time.sleep(self.sleep_for)


def test_timeit_no_metadata_err(proj_path):
    with pytest.raises(AttributeError):
        SleepClassNoMetadata().run()


@pytest.mark.parametrize("eager", [True, False])
def test_timeit(proj_path, eager):
    with zntrack.Project() as project:
        sleep_class = SleepClass()

    project.run(eager=eager)
    # if not eager:
    #     sleep_class.load()
    assert pytest.approx(sleep_class.timeit_metrics["run"], 0.1) == 1.0


@pytest.mark.parametrize("eager", [True, False])
def test_timeit_loop(proj_path, eager):
    with zntrack.Project() as project:
        sleep_class = SleepClassLoop()
    project.run(eager=eager)
    # if not eager:
    #     sleep_class.load()

    assert pytest.approx(sleep_class.timeit_metrics["sleep"]["mean"], 0.1) == 0.1
    assert sleep_class.timeit_metrics["sleep"]["std"] < 1e-2
    assert len(sleep_class.timeit_metrics["sleep"]["values"]) == 30



"""Module containing functions to finalize an experiment."""

import git


def make_commit(msg: str, path: str = ".") -> str:
    """Create a new GIT commit.

    Parameters
    ----------
    msg : str
        Commit message.
    path : str, optional
        Path to the GIT repository, by default "."

    Returns
    -------
    str
        The hash of the new commit.

    """
    if msg is None:
        msg = "zntrack: auto commit"
    repo = git.Repo(path)
    if repo.is_dirty():
        repo.git.add(".")
        repo.git.commit("-m", msg)
    return repo.head.commit.hexsha
